import datetime as dt
import os
import numpy as np
import pandas as pd
import matplotlib.dates as mdates
from astropy.time import Time
from scipy.constants import speed_of_light
from scipy.ndimage import uniform_filter1d
from numpy.polynomial import Polynomial
from matplotlib import font_manager
import georinex as gr
import sys
import time
from pyOASIS import settings
from pyOASIS import linear_combinations
from pyOASIS import gnss_freqs
import warnings
import pyOASIS

def RNXclean(station_name,day_of_year,year,input_folder,orbit_folder,output_folder):

    # Accessing the frequencies of the GPS system
    gps_freqs = gnss_freqs.FREQUENCY[gnss_freqs.GPS]
    f1 = gps_freqs[1]
    f2 = gps_freqs[2]
    f5 = gps_freqs[5]
    
    # Calculating the frequencies of the GLONASS system (glonass_channels.dat)
    # To obtain the GLONASS frequencies (code 'R'):
    file_name = os.path.join(pyOASIS.__path__[0], 'glonass_channels.dat')
    
    # Read the file into a DataFrame with column names defined
    df_slots = pd.read_csv(file_name, sep=' ', header=None, names=['Slot', 'Channel'])
    glonass_frequencies = gnss_freqs.FREQUENCY[gnss_freqs.GLO]
    
    # List to store the data
    data = []
    
    # Iterate over each row of the DataFrame
    for index, row in df_slots.iterrows():
        satellite = row['Slot']
        k = row['Channel']
        row_data = [satellite]
    
        for channel, frequency in glonass_frequencies.items():
            if callable(frequency):  # Check if it is a lambda function
                freq_value = frequency(k)
            else:
                freq_value = frequency
            formatted_freq = f"{freq_value:.1f}"
            row_data.append(formatted_freq)
        data.append(row_data)
    
    # Convert the list of lists into a pandas DataFrame
    glo_freqs = pd.DataFrame(data, columns=['Satellite', 'fr1', 'fr2', 'fr3'])
    
    # Initial time, number of hours, and xticks interval on the graph.
    h1 = 0
    n_horas = 24
    int1 = 120
    
    # Minimum number of observations per arc
    ARCL = 15
    
    constellations = ['G', 'R']  # G: GPS, R: GLONASS, E: Galileo (future), C: BeiDou (future)
    
    for c in constellations:
        sat_class = c
    
        # RINEX OPENING:
        # ------------------------------------
        # Selection of version for RINEX files
        version_number = '1'
        print()
        year_format = f"{year[-2:]}o"
        rinex_file_path = f"{input_folder}/{station_name.lower()}{day_of_year}{version_number}.{year_format}"
        if not os.path.exists(rinex_file_path):
            version_number = '0' if version_number == '1' else '1'
            rinex_file_path = f"{input_folder}/{station_name.lower()}{day_of_year}{version_number}.{year_format}"
            if not os.path.exists(rinex_file_path):
                print(f"File {rinex_file_path} not found either. Please check the path or other issues.")
            else:
                print(f"File {rinex_file_path} found with opposite version of {version_number}.")
        else:
            print(f"File {rinex_file_path} found with version_number = {version_number}")
    
        # Supressing warning about timedelta
        warnings.filterwarnings("ignore", message="Converting non-nanosecond precision datetime values to nanosecond precision.", category=UserWarning)
        #obs_data = gr.load(rinex_file_path)


        try:
            # Tenta carregar normalmente
            obs_data = gr.load(rinex_file_path)
        except IndexError as e:
            print(f"[WARNING] IndexError when loading: {rinex_file_path}")
            print("[INFO] Retrying with only GPS satellites...")
            # Fallback: só GPS
            obs_data = gr.load(rinex_file_path, use='G')


        print()
    
        # ORBIT FILE OPENING:
        # -------------------------
        # Path and time definitions
    
        freq = int(obs_data.interval)
        file_path = os.path.join(orbit_folder, f'ORBITS_{year}_{day_of_year}.SP3')
        column_names = ["Date", "Time", "Satélite", "X", "Y", "Z"]
        df2 = pd.read_csv(file_path, sep="\t", header=0, names=column_names)
    
        # List to store IPP coordinates
        result = {"Date": [], "Time": [], "SAT": [], "lon": [], "lat": [], "el": []}
        index = []
    
        # Filtering satellites by class
        satellites_to_plot = [sv for sv in np.unique(obs_data.sv.values) if sv.startswith(sat_class)]
    
        # List to store DataFrames for each satellite
        dfs = []
    
        # Remove the first character of each element in the 'Satellite' column
        df2['Satélite'] = df2['Satélite'].str[1:]
    
        IPP_UNIQ = np.unique(df2['Satélite'])
    
        # Converting lists to sets
        set_IPP_UNIQ = set(IPP_UNIQ)
        set_satellites_to_plot = set(satellites_to_plot)
    
        # Finding the intersection between the two sets
        common_elements = set_IPP_UNIQ.intersection(set_satellites_to_plot)
    
        # Converting the set of common elements back into a list, if necessary
        common_elements_list = list(common_elements)
    
        # Defining the desired order for the initial letters
        order = {'G': 1, 'R': 2, 'E': 3, 'C': 4}
    
        # Sorting the list based on the initial letter and the last two digits
        sorted_common_elements_list = sorted(common_elements_list, key=lambda x: (order[x[0]], int(x[-2:])))
    
        for sat in sorted_common_elements_list:
            print()
            print("Processing satellite:", sat)
            print()
    
            # Checking the satellite class and adjusting the values of f1, f2, f5
            if sat_class == 'G':
                f1 = f1
                f2 = f2
                f5 = f5
            elif sat_class == 'R':
                # Locating the row where 'Satellite' matches 'sat'
                sat_row = glo_freqs.loc[glo_freqs['Satellite'] == sat]
    
                if not sat_row.empty:
                    f1 = float(sat_row['fr1'].values[0])
                    f2 = float(sat_row['fr2'].values[0])
                    f5 = float(sat_row['fr3'].values[0])
                else:
                    f1 = f2 = f5 = None  # Or default values

            L1 = np.array(obs_data['L1'].sel(sv=sat))
            L2 = np.array(obs_data['L2'].sel(sv=sat))
    
            # Adding the L5 frequency to L1 and L2
            if 'L5' in obs_data:
                L5 = np.array(obs_data['L5'].sel(sv=sat))
            else:
                L5 = np.full_like(L1, np.nan)  # Fill with NaN if L5 data is not available
    
            if 'P1' in obs_data:
                P1 = np.array(obs_data['P1'].sel(sv=sat))
                code_obs1 = 'P1'
            else:
                P1 = np.array(obs_data['C1'].sel(sv=sat))
                code_obs1 = 'C1'
            if np.all(np.isnan(P1)):
                P1 = np.array(obs_data['C1'].sel(sv=sat))
                code_obs1 = 'C1'
            else:
                code_obs1 = 'P1'
            if 'P2' in obs_data:
                P2 = np.array(obs_data['P2'].sel(sv=sat))
                code_obs2 = 'P2'
            else:
                P2 = np.array(obs_data['C2'].sel(sv=sat))
                code_obs2 = 'C2'
            if np.all(np.isnan(P2)):
                P2 = np.array(obs_data['C2'].sel(sv=sat))
                code_obs2 = 'C2'
            else:
                code_obs2 = 'P2'
    
            # Checking for the presence of carrier phase (P5) or pseudorange (C5) observations
            if 'P5' in obs_data:
                P5 = np.array(obs_data['P5'].sel(sv=sat))
                code_obs5 = 'P5'
                if np.all(np.isnan(P5)):
                    P5 = np.array(obs_data['C5'].sel(sv=sat))
                    code_obs5 = 'C5'
                    if np.all(np.isnan(P5)):
                        # print(f"WARNING: All values in {code_obs5} are NaNs for satellite {sat}")
                        print()
            elif 'C5' in obs_data:
                P5 = np.array(obs_data['C5'].sel(sv=sat))
                code_obs5 = 'C5'
                if np.all(np.isnan(P5)):
                    # print(f"WARNING: All values in {code_obs5} are NaNs for satellite {sat}")
                    print()
            else:
                # If neither P5 nor C5 are present, set P5 as a column of NaNs
                P5 = np.full_like(L5, np.nan)
                code_obs5 = "None"  # Or any other value to indicate the absence of data, as you prefer
    
            if np.all(np.isnan(P5)):
                print(f"WARNING: All values in L5 data", f"{code_obs5} are NaNs for satellite {sat}")
    
            df = pd.DataFrame({'time': obs_data.time})
            df.set_index('time', inplace=True)
            df['index'] = df.index
            df['mjd'] = Time(df.index).mjd
            mjd = Time(df.index).mjd
    
            df['timestamp'] = pd.to_datetime(df.index)
    
            # Extracting only the time from 'timestamp' and storing it in a new 'time' column
            df['time'] = df['timestamp'].dt.time
            df['date'] = df['timestamp'].dt.date
    
            df['L1'] = L1
            df['L2'] = L2
            df['L5'] = L5
            df['P1'] = P1
            df['P2'] = P2
            df['P5'] = P5
    
            df['satellite'] = [sat] * len(L1)
            df['station'] = [station_name.upper()] * len(L1)
            df['position'] = [obs_data.position] * len(L1)
            df['mjd'] = ["{:.12f}".format(valor) for valor in mjd]
    
            timep = df['time']
    
            L1, L2, L5, P1, P2, P5 = df['L1'] , df['L2'] , df['L5'] , df['P1'] , df['P2'], df['P5']
    
            mjd, date, time = df['mjd'] , df['date'] , df['time']
    
            df['LMW12'] = linear_combinations.melbourne_wubbena(f1, f2, L1, L2, P1, P2)
    
            df['LMW15'] = linear_combinations.melbourne_wubbena(f1, f5, L1, L5, P1, P5)
    
            station_coords = df['position'].iloc[0]
    
            # Converting each string in the list into a float (or int, if appropriate)
            coords_list = [float(coord) for coord in station_coords]
    
            # Converting the list of coordinates into a list containing a single tuple
            obs_x, obs_y, obs_z = coords_list[0], coords_list[1], coords_list[2]
    
            # Initializing empty lists to collect data
            all_data = []
            all_time = []
            all_sat = []
            all_longitude = []
            all_latitude = []
            all_elevation = []
    
            # for ipp_sat in sorted_common_elements_list:
            indices = np.where(df2['Satélite'] == sat)[0]
            df_filtrado = df2.iloc[indices]
    
            import time

            # Getting the current time in seconds since the Unix Epoch
            current_time = time.time()
    
            # Converting the time into a local time structure
            local_time = time.localtime(current_time)
    
            # Formatting the time in hh:mm format
            formatted_time = time.strftime("%H:%M:%S", local_time)

            from datetime import datetime, timedelta
    
            # Defining the interpolation rate in seconds (e.g., 15 seconds)
            rate = freq

            # Lists to store the original and interpolated data
            all_data = []
            all_time = []
            all_sat = []
            all_longitude = []
            all_latitude = []
            all_elevation = []
    
            # Variables to store the last known data for each satellite
            last_data_by_satellite = {}
    
            for _, row in df_filtrado.iterrows():
                date = row["Date"]
                sat = row["Satélite"]
                sx = row["X"]
                sy = row["Y"]
                sz = row["Z"]
                time = row["Time"]
    
                # Assuming that convert_coords and IonosphericPiercingPoint are functions defined earlier
                lon, lat, alt = settings.convert_coords(obs_x, obs_y, obs_z, to_radians=True)
                ip = settings.IonosphericPiercingPoint(sx, sy, sz, obs_x, obs_y, obs_z)
                elevation = ip.elevation(lat, lon)
                lat_ip, lon_ip = ip.coordinates(lat, lon)
    
                # Converting the current date and time into a datetime object to calculate time intervals
                current_time = datetime.strptime(f"{date} {time}", "%d-%m-%Y %H:%M:%S")
    
                # If we already have previous data for the satellite, we check whether interpolation is needed
                if sat in last_data_by_satellite:
                    last_known_time = last_data_by_satellite[sat]['date_time']
    
                    # If the difference between the last known time and the current time is greater than the rate, interpolate
                    while last_known_time + timedelta(seconds=rate) < current_time:
                        last_known_time += timedelta(seconds=rate)
    
                        # Add the interpolated data to the lists (using the last known values)
                        all_data.append(last_known_time.strftime("%d-%m-%Y"))
                        all_time.append(last_known_time.strftime("%H:%M:%S"))
                        all_sat.append(sat)
                        all_longitude.append(last_data_by_satellite[sat]['lon'])
                        all_latitude.append(last_data_by_satellite[sat]['lat'])
                        all_elevation.append(last_data_by_satellite[sat]['elevation'])
    
                # Update the satellite data for the next cycle
                last_data_by_satellite[sat] = {
                    'date_time': current_time,
                    'lon': lon_ip,
                    'lat': lat_ip,
                    'elevation': elevation
                }
    
                # Add the original data (without interpolation) to the lists
                all_data.append(date)
                all_time.append(time)
                all_sat.append(sat)
                all_longitude.append(lon_ip)
                all_latitude.append(lat_ip)
                all_elevation.append(elevation)
    
            # After processing all satellites:
            # Set the final time as 23:59:45
            final_hour = datetime.strptime(f"{date} 23:59:45", "%d-%m-%Y %H:%M:%S")
    
            # Iterate over each satellite to extrapolate the data until the end of the day
            for satellite_name, data in last_data_by_satellite.items():
                last_record = data  # Last recorded data for the satellite
                while last_record['date_time'] < final_hour:
                    last_record['date_time'] += timedelta(seconds=rate)  # Increment by 15 seconds
    
                    # Add the extrapolated data (the last known values) to the lists
                    all_data.append(last_record['date_time'].strftime("%d-%m-%Y"))
                    all_time.append(last_record['date_time'].strftime("%H:%M:%S"))
                    all_sat.append(satellite_name)
                    all_longitude.append(last_record['lon'])
                    all_latitude.append(last_record['lat'])
                    all_elevation.append(last_record['elevation'])
    
            # After exiting the loop, convert the lists to NumPy arrays (if necessary)
            all_date = pd.Series(all_data)
            all_time = pd.Series(all_time)
            all_sat = pd.Series(all_sat)
            all_longitude = pd.Series(all_longitude)
            all_latitude = pd.Series(all_latitude)
            all_elevation = pd.Series(all_elevation)
    
            # Create a new combined DataFrame with the original, interpolated, and extrapolated data
            combined_df = pd.DataFrame({
                "Date": all_date,
                "Time": all_time,
                "SAT": all_sat,
                "Longitude": all_longitude,
                "Latitude": all_latitude,
                "Elevation": all_elevation
            })
    
            # Converting the time data to datetime type
            combined_df['Time'] = pd.to_datetime(combined_df['Time'], format='%H:%M:%S')
    
            df['time'] = df['time'].astype(str)
    
            # Converting the times into datetime objects
            combined_df['Time'] = pd.to_datetime(combined_df['Time']).dt.time
    
            # Specify the format when converting to datetime
            df['time'] = pd.to_datetime(df['time'], format='%H:%M:%S').dt.time
    
            # Finding the times that are present in both dataframes
            horas_em_common = set(combined_df['Time']).intersection(set(df['time']))
    
            # Defining the reference length as the smaller length between the two dataframes
            comprimento_referencia = min(len(combined_df), len(df))
    
            # Filtering the dataframes to keep only the common rows based on the reference length
            combined_df = combined_df[combined_df['Time'].isin(horas_em_common)].iloc[:comprimento_referencia]
            df = df[df['time'].isin(horas_em_common)].iloc[:comprimento_referencia]
    
            colunas_desejadas_df1 = ['Elevation','Longitude','Latitude']
    
            combined_df_selecionado = combined_df[colunas_desejadas_df1]
    
            L1, L2, L5, P1, P2, P5 = df['L1'] , df['L2'] , df['L5'], df['P1'] , df['P2'] , df['P5']
    
            mjd, date, time = df['mjd'] , df['date'] , df['time']
    
            # Assuming that df_filtered is your DataFrame
            # Split the 'position' column into three new columns and expand the result into separate columns
            df['pos_x'], df['pos_y'], df['pos_z'] = obs_x, obs_y, obs_z
    
            df['height'] = np.full(len(L1), 450.0)
    
            # Now df_filtered has the columns 'pos_x', 'pos_y', and 'pos_z'
            # You can then select these new columns for inclusion in the DataFrame to be saved
            colunas_desejadas_df2 = ['date','time', 'mjd', 'pos_x', 'pos_y', 'pos_z', 'L1', 'L2', 'L5', 'P1', 'P2', 'P5', 'satellite','station','height']
            df_filtered_selecionado = df[colunas_desejadas_df2]
    
            # Make sure that df_filtered and combined_df have the same length
            if len(df_filtered_selecionado) == len(combined_df_selecionado):
                # Merging the DataFrames side by side
                df_final = pd.concat([df_filtered_selecionado, combined_df_selecionado], axis=1)
            else:
                print("Error: df_filtered and combined_df do not have the same length!")
                # Exit the execution if the DataFrames do not have the same size
                sys.exit()
    
            LMW2 = df['LMW12']
            LMW3 = df['LMW15']
    
            abs_elevation = abs(combined_df_selecionado['Elevation'])
    
            df['LMW2'] = LMW2
            df['LMW3'] = LMW3
    
            combined_df_selecionado.index = df.index
    
            indices_low_elevation = combined_df_selecionado.index[abs_elevation < 10]
    
            # List of columns that should be set to NaN
            cols_nan = ['LMW2', 'LMW3']
    
            # Assign NaN to the specified columns only for the rows in indices_low_elevation
            df.loc[indices_low_elevation, cols_nan] = np.nan
    
            # Assuming that LMW is a NumPy array and time is a corresponding time vector
            LMW = np.array(df['LMW2'])  # Ensure that LMW is a NumPy array to facilitate manipulation
            LMW15 = np.array(df['LMW3'])  # Ensure that LMW is a NumPy array to facilitate manipulation
    
            arcs = []  # List to store the observation arcs
            current_arc = []  # Temporary list to store the current observation arc
    
            # Iterate over all elements of LMW
            for idx, value in enumerate(LMW):
                if np.isnan(value):
                    # If the current value is NaN, check if the current arc is empty
                    # This avoids adding empty arcs in case of consecutive NaNs
                    if current_arc:
                        arcs.append(current_arc)
                        current_arc = []
                else:
                    # If the value is not NaN, add the index to the current arc
                    current_arc.append(idx)
    
            # Add the last arc if it is not empty
            if current_arc:
                arcs.append(current_arc)
    
            print()
            print('Melbourne-Wubbena combination for L1-L2')
            print()
    
            # Print information about each arc and classify them
            for i, arc in enumerate(arcs):
                start_index = arc[0]
                end_index = arc[-1]
                num_observations = len(arc)
                status = "Kept" if num_observations >= 15 else "Discarded"
    
                print(f"Arc {i + 1}: Start index = {start_index}, End index = {end_index}, "
                    f"Number of observations = {num_observations}, Status = {status}")
    
            arcs15 = []  # List to store the observation arcs
            current_arc15 = []  # Temporary list to store the current observation arc
    
            # Iterate over all elements of LMW15
            for idx, value in enumerate(LMW15):
                if np.isnan(value):
                    # If the current value is NaN, check if the current arc is empty
                    # This avoids adding empty arcs in case of consecutive NaNs
                    if current_arc15:
                        arcs15.append(current_arc15)
                        current_arc15 = []
                else:
                    # If the value is not NaN, add the index to the current arc
                    current_arc15.append(idx)
    
            # Add the last arc if it is not empty
            if current_arc15:
                arcs15.append(current_arc15)
    
            print()
            print()
    
            print('Melbourne-Wubbena combination for L1-L5')
    
            print()
    
            # Print information about each arc and classify them
            for i, arc in enumerate(arcs15):
                start_index = arc[0]
                end_index = arc[-1]
                num_observations = len(arc)
                status = "Kept" if num_observations >= 15 else "Discarded"
                print(f"Arc {i + 1}: Start index = {start_index}, End index = {end_index}, "
                    f"Number of observations = {num_observations}, Status = {status}")
    
            LMW = np.array(LMW)  # Still assuming that LMW is your data array
            LMW15 = np.array(LMW15)  # Still assuming that LMW is your data array
    
            print()
            print()
    
            # Function to rescale the data to the range [-10, 10]
            def rescale_data(data):
                min_val = np.min(data)
                max_val = np.max(data)
                # Rescale the data to the range [0, 1]
                scaled_data = (data - min_val) / (max_val - min_val)
                # Adjust it to the range [-10, 10]
                final_data = scaled_data * 20 - 10
                return final_data
    
    
            # -- [L1 - L2]
            idx_total = []
    
            for arc in arcs:
                arc_data = df.iloc[arc]
                time = df.index[arc]
    
                if len(arc_data) < ARCL:  # Check if there are enough points for the fitting
                    continue
    
                x = arc_data.index.astype(np.int64) // 10**9  # Converting to seconds
                xx = arc_data['time']
                y = arc_data['LMW2'].values
    
                y_rescaled = rescale_data(y)
                delta_y = np.diff(y_rescaled, prepend=np.nan)
    
                # Fit a polynomial only to the valid values (excluding np.nan)
                p = Polynomial.fit(x[1:], delta_y[1:], 3)
    
                delta_y_fit = p(x)  # Polynomial fitted values
                residuals = abs(delta_y - delta_y_fit)  # Calculate residuals
    
                # Calculation of customized quartiles and IQR for outlier identification
                Q1 = np.nanpercentile(residuals, 15)
                Q3 = np.nanpercentile(residuals, 85)
                IQR = Q3 - Q1
                outlier_mask = (residuals < Q1 - 4 * IQR) | (residuals > Q3 + 4 * IQR)
                high_residuals_mask = residuals > 1  # Máscara para resíduos altos
                other_residuals_mask = ~(outlier_mask | high_residuals_mask)  # Máscara para os demais resíduos

                # Suppose that 'residuals' is the variable containing the calculated residuals
                limiar_outlier = 4  # Threshold for identifying outliers (e.g., 4 times the IQR)
                limiar_residuo_alto = 1  # Threshold for identifying high residuals
    
                # Identify indices of outliers and high residuals
                indices_outliers =  arc[0] + np.where(np.abs(residuals) > limiar_outlier * IQR)[0]
                indices_residuos_altos = arc[0] + np.where(np.abs(residuals) > limiar_residuo_alto)[0]
    
                # Merge the indices without duplicates
                indices_combinados = np.union1d(indices_outliers, indices_residuos_altos)
    
                # Add the combined_indices values to the list
                idx_total.append(indices_combinados)
    
            # Concatenate all values in the list into a single NumPy array
            if idx_total:
                idx_total = np.concatenate(idx_total)
    
            # Initialize 'flags' with 'C' for all indices in LMW
            flags = ['C'] * len(LMW)
    
            # Check where LMW is NaN and replace the corresponding flags with 'S'
            nan_indices = np.where(np.isnan(LMW))[0]
            flags = list(flags)  # Convert flags to a list
            for idx in nan_indices:
                flags[idx] = 'S'
    
            # Use idx_total as indices to change the corresponding flags to 'S'
            for idx in idx_total:
                if idx < len(flags):
                    flags[idx] = 'S'
    
            for idx in idx_total:
                print("L1-L2 (+)", idx, timep.iloc[idx])
    
    
            # ---- [L1 - L5]
            idx_total15 = []
    
            for arc in arcs15:
                arc_data = df.iloc[arc]
                time = df.index[arc]
    
                if len(arc_data) < ARCL:  # Check if there are enough points for fitting
                    continue
    
                x = arc_data.index.astype(np.int64) // 10**9  # Convert to seconds
                xx = arc_data['time']
                y = arc_data['LMW3'].values
    
                y_rescaled = rescale_data(y)
                delta_y = np.diff(y_rescaled, prepend=np.nan)
    
                # Fit a polynomial only to valid values (excluding np.nan)
                p = Polynomial.fit(x[1:], delta_y[1:], 3)

                delta_y_fit = p(x)  # Polynomial fitted values
                residuals = abs(delta_y - delta_y_fit)  # Calculate residuals
    
                # Calculation of customized quartiles and IQR for outlier identification
                Q1 = np.nanpercentile(residuals, 15)
                Q3 = np.nanpercentile(residuals, 85)
                IQR = Q3 - Q1
                outlier_mask = (residuals < Q1 - 4 * IQR) | (residuals > Q3 + 4 * IQR)
                high_residuals_mask = residuals > 1  # Mask for high residuals
                other_residuals_mask = ~(outlier_mask | high_residuals_mask)  # Mask for normal residuals
    
                # Assume that 'residuals' is the variable containing the calculated residuals
                limiar_outlier = 4  # Threshold to identify outliers (e.g., 4 times the IQR)
                limiar_residuo_alto = 1  # Threshold to identify high residuals
    
                # Identify indices of outliers and high residuals
                indices_outliers15 =  arc[0] + np.where(np.abs(residuals) > limiar_outlier * IQR)[0]
                indices_residuos_altos15 = arc[0] + np.where(np.abs(residuals) > limiar_residuo_alto)[0]
    
                # Combine the indices without duplication
                indices_combinados15 = np.union1d(indices_outliers15, indices_residuos_altos15)
    
                # Add the combined indices to the list
                idx_total15.append(indices_combinados15)
    
                print("Combined indices (L1 - L5):", indices_combinados15)
    
            # Concatenate all values from the list into a single NumPy array
            if idx_total15:
                idx_total15 = np.concatenate(idx_total15)
    
            # Check if LMW15 is not entirely composed of NaN
            if not np.all(np.isnan(LMW15)):
                # Identify where LMW15 is NaN and assign 'S' to the corresponding flags
                nan_indices15 = np.where(np.isnan(LMW15))[0]
                for idx in nan_indices15:
                    flags[idx] = 'S'
    
            # Use idx_total15 as indices to set the corresponding flags to 'S'
            for idx in idx_total15:
                if idx < len(flags):
                    flags[idx] = 'S'
    
            for idx in idx_total15:
                print("L1-L5 (+)", idx, timep.iloc[idx])
    
            satellite_values = [sat] * len(L1)
            station = [station_name.upper()] * len(L1)
            position = [obs_data.position] * len(L1)
    
            L1 = [value if pd.notna(value) else -999999.999 for value in L1]
            L2 = [value if pd.notna(value) else -999999.999 for value in L2]
            L5 = [value if pd.notna(value) else -999999.999 for value in L5]
    
            P1 = [value if pd.notna(value) else -999999.999 for value in P1]
            P2 = [value if pd.notna(value) else -999999.999 for value in P2]
            P5 = [value if pd.notna(value) else -999999.999 for value in P5]
    
            L1 = ["{:.3f}".format(valor) for valor in L1]
            L2 = ["{:.3f}".format(valor) for valor in L2]
            L5 = ["{:.3f}".format(valor) for valor in L5]
    
            P1 = ["{:.3f}".format(valor) for valor in P1]
            P2 = ["{:.3f}".format(valor) for valor in P2]
            P5 = ["{:.3f}".format(valor) for valor in P5]
    
            df.reset_index(drop=True, inplace=True)
            combined_df.reset_index(drop=True, inplace=True)
    
            export_df = pd.DataFrame({
                'date': df['date'],
                'time': df['time'],
                'mjd': df['mjd'],
                'pos_x': df['pos_x'],
                'pos_y': df['pos_y'],
                'pos_z': df['pos_z'],
                'L1': df['L1'],
                'L2': df['L2'],
                'L5': df['L5'],
                'P1': df['P1'],
                'P2': df['P2'],
                'P5': df['P5'],
                'cs_flags': flags,
                'satellite': satellite_values,
                'sta': station,
                'hght': df['height'],
                'El': abs(combined_df['Elevation'].round(2)),
                'Lon': combined_df['Longitude'],
                'Lat': combined_df['Latitude'],
            })
    
            export_df.rename(columns={'P1': code_obs1}, inplace=True)
            export_df.rename(columns={'P2': code_obs2}, inplace=True)
            export_df.rename(columns={'P5': code_obs5}, inplace=True)
    
            file_name = f"{station_name.upper()}_{sat}_{day_of_year}_{year}.RNX1"
    
            output_directory = os.path.join(output_folder)
            full_path = output_directory
    
    
            os.makedirs(full_path, exist_ok=True)
            output_file_path = os.path.join(full_path, file_name)
            export_df.to_csv(output_file_path, sep='\t', index=False, na_rep='-999999.999')



import os
import numpy as np
import sys
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from astropy.time import Time
from scipy.constants import speed_of_light
from pyOASIS import gnss_freqs
from pyOASIS import levelling_settings
import pyOASIS

def RNXlevelling(estacao, diretorio_principal, show_plot=True):

    # Suppress all error and warning messages by redirecting stderr to the system null device
    sys.stderr = open(os.devnull, 'w')

    # Variables and Parameters
    h1 = 0
    n_horas = 24  # hours
    int1 = 120  # minutes

    # Accessing the frequencies of the GPS system
    gps_freqs = gnss_freqs.FREQUENCY[gnss_freqs.GPS]
    f1 = gps_freqs[1]
    f2 = gps_freqs[2]
    f5 = gps_freqs[5]

    # Define the file name
    file_name = pyOASIS.__path__[0] + '/glonass_channels.dat'  # option without slashes

    # Read the file into a DataFrame with defined column names
    df_slots = pd.read_csv(file_name, sep=' ', header=None, names=['Slot', 'Channel'])
    glonass_frequencies = gnss_freqs.FREQUENCY[gnss_freqs.GLO]

    # List to store the data
    data = []

    # Iterate over each row in the DataFrame
    for index, row in df_slots.iterrows():
        satellite = row['Slot']
        k = row['Channel']
        row_data = [satellite]

        for channel, frequency in glonass_frequencies.items():
            if callable(frequency):  # Check if it's a lambda function
                freq_value = frequency(k)
            else:
                freq_value = frequency
            formatted_freq = f"{freq_value:.1f}"
            row_data.append(formatted_freq)

        # Add row data to the data list
        data.append(row_data)

    # Convert the list of lists into a pandas DataFrame
    glo_freqs = pd.DataFrame(data, columns=['Satellite', 'fr1', 'fr2', 'fr3'])

    # Building the full path to the folder
    caminho_ = os.path.join(diretorio_principal)

    # Checking if the directory exists
    if os.path.exists(caminho_):
        conteudo_ = os.listdir(caminho_)

        # Defining the 'arquivos' variable to include only files ending with .RNX2
        arquivos = [arquivo for arquivo in conteudo_ if arquivo.endswith(".RNX2")]

        first = arquivos[0]
        doy = first[9:12]
        ano = first[13:17]

        # Sorting files by satellite number
        arquivos_ordenados = sorted(arquivos, key=lambda x: int(x.split("_")[1][1:]))

        # Printing the found files
        for arquivo in arquivos_ordenados:
            print("File:", arquivo)

        print()
        # Counting the number of files
        numero_de_arquivos = len(arquivos_ordenados)
        print("Number of RINEX_SCREENED (.RNX2) files in the directory:", numero_de_arquivos)

    else:
        print("The specified directory does not exist.")

    print()

    # Initializing lists to store values from each variable of all files
    date = []
    time = []
    mjd = []
    pos_x = []
    pos_y = []
    pos_z = []
    L1 = []
    L2 = []
    L5 = []
    P1 = []
    P2 = []
    P5 = []
    cs_flag = []
    outlier_flag = []
    satellites = []
    sta = []
    hght = []
    El = []
    Lon = []
    Lat = []
    obs_La = []
    obs_Lb = []
    obs_Lc = []
    obs_Ca = []
    obs_Cb = []
    obs_Cc = []

    for arquivo in arquivos:
        caminho_arquivo = os.path.join(caminho_, arquivo)

        with open(caminho_arquivo, 'r') as f:
            header = f.readline().strip().split('\t')

            for linha in f:
                colunas = linha.strip().split('\t')
                registro = {
                    'date': colunas[0],
                    'time': colunas[1],
                    'mjd': colunas[2],
                    'pos_x': colunas[3],
                    'pos_y': colunas[4],
                    'pos_z': colunas[5],
                    'L1': colunas[6],
                    'L2': colunas[7],
                    'L5': colunas[8],
                    'P1': colunas[9],
                    'P2': colunas[10],
                    'P5': colunas[11],
                    'cs_flag': colunas[12],
                    'outlier_flag': colunas[13],
                    'satellite': colunas[14],
                    'sta': colunas[15],
                    'hght': colunas[16],
                    'El': colunas[17],
                    'Lon': colunas[18],
                    'Lat': colunas[19],
                    'obs_La': colunas[20],
                    'obs_Lb': colunas[21],
                    'obs_Lc': colunas[22],
                    'obs_Ca': colunas[23],
                    'obs_Cb': colunas[24],
                    'obs_Cc': colunas[25]
                }

                # Appending each variable to its respective list
                date.append(registro['date'])
                time.append(registro['time'])
                mjd.append(registro['mjd'])
                pos_x.append(registro['pos_x'])
                pos_y.append(registro['pos_y'])
                pos_z.append(registro['pos_z'])
                L1.append(registro['L1'])
                L2.append(registro['L2'])
                L5.append(registro['L5'])
                P1.append(registro['P1'])
                P2.append(registro['P2'])
                P5.append(registro['P5'])
                cs_flag.append(registro['cs_flag'])
                outlier_flag.append(registro['outlier_flag'])
                satellites.append(registro['satellite'])
                sta.append(registro['sta'])
                hght.append(registro['hght'])
                El.append(registro['El'])
                Lon.append(registro['Lon'])
                Lat.append(registro['Lat'])
                obs_La.append(registro['obs_La'])
                obs_Lb.append(registro['obs_Lb'])
                obs_Lc.append(registro['obs_Lc'])
                obs_Ca.append(registro['obs_Ca'])
                obs_Cb.append(registro['obs_Cb'])
                obs_Cc.append(registro['obs_Cc'])

    sat_classes = ['G', 'R']

    # Create a single figure
    plt.figure(figsize=(12, 6))

    # Iterating over sat_class values
    for sat_class in sat_classes:
        sat = sat_class

        if sat:
            satellites_to_plot = [sv for sv in np.unique(satellites) if sv.startswith(sat)]
        else:
            satellites_to_plot = np.unique(satellites)

        # Initializing lists to store the values of each variable from all files
        date_filtered = []
        time_filtered = []
        mjd_filtered = []
        pos_x_filtered = []
        pos_y_filtered = []
        pos_z_filtered = []
        L1_filtered = []
        L2_filtered = []
        L5_filtered = []
        P1_filtered = []
        P2_filtered = []
        P5_filtered = []
        cs_flag_filtered = []
        outlier_flag_filtered = []
        satellite_filtered = []
        sta_filtered = []
        hght_filtered = []
        El_filtered = []
        Lon_filtered = []
        Lat_filtered = []
        obs_La_filtered = []
        obs_Lb_filtered = []
        obs_Lc_filtered = []
        obs_Ca_filtered = []
        obs_Cb_filtered = []
        obs_Cc_filtered = []

        for satellite in satellites_to_plot:
            print()
            print(f"Processing {satellite} satellite...")
            sat = satellite

            # Checking satellite class and adjusting f1, f2, f5 values
            if sat.startswith('G'):
                f1 = f1
                f2 = f2
                f5 = f5
            elif sat.startswith('R'):
                # Locate the row where 'Satellite' matches 'sat'
                sat_row = glo_freqs.loc[glo_freqs['Satellite'] == sat]

                if not sat_row.empty:
                    f1 = float(sat_row['fr1'].values[0])
                    f2 = float(sat_row['fr2'].values[0])
                    f5 = float(sat_row['fr3'].values[0])
            else:
                f1 = f2 = f5 = None  # Or default values

            akl = 40.3 * 10 ** 16 * ((1 / f2 ** 2) - (1 / f1 ** 2))

            lambda1 = (speed_of_light / f1)
            lambda2 = (speed_of_light / f2)
            lambda5 = (speed_of_light / f5)

            indices = np.where(np.array(satellites) == satellite)[0]

            # Initializing filtered lists for each satellite
            date_filtered = []
            time_filtered = []
            mjd_filtered = []
            pos_x_filtered = []
            pos_y_filtered = []
            pos_z_filtered = []
            L1_filtered = []
            L2_filtered = []
            L5_filtered = []
            P1_filtered = []
            P2_filtered = []
            P5_filtered = []
            cs_flag_filtered = []
            outlier_flag_filtered = []
            satellite_filtered = []
            sta_filtered = []
            hght_filtered = []
            El_filtered = []
            Lon_filtered = []
            Lat_filtered = []
            obs_La_filtered = []
            obs_Lb_filtered = []
            obs_Lc_filtered = []
            obs_Ca_filtered = []
            obs_Cb_filtered = []
            obs_Cc_filtered = []

            for idx in indices:
                date_filtered.append(date[idx])
                time_filtered.append(time[idx])
                mjd_filtered.append(mjd[idx])
                pos_x_filtered.append(pos_x[idx])
                pos_y_filtered.append(pos_y[idx])
                pos_z_filtered.append(pos_z[idx])
                L1_filtered.append(L1[idx])
                L2_filtered.append(L2[idx])
                L5_filtered.append(L5[idx])
                P1_filtered.append(P1[idx])
                P2_filtered.append(P2[idx])
                P5_filtered.append(P5[idx])
                cs_flag_filtered.append(cs_flag[idx])
                outlier_flag_filtered.append(outlier_flag[idx])
                satellite_filtered.append(satellites[idx])
                sta_filtered.append(sta[idx])
                hght_filtered.append(hght[idx])
                El_filtered.append(El[idx])
                Lon_filtered.append(Lon[idx])
                Lat_filtered.append(Lat[idx])
                obs_La_filtered.append(obs_La[idx])
                obs_Lb_filtered.append(obs_Lb[idx])
                obs_Lc_filtered.append(obs_Lc[idx])
                obs_Ca_filtered.append(obs_Ca[idx])
                obs_Cb_filtered.append(obs_Cb[idx])
                obs_Cc_filtered.append(obs_Cc[idx])

            # Constructing a DataFrame with the filtered lists
            data = {
                'date': date_filtered,
                'time2': time_filtered,
                'mjd': mjd_filtered,
                'pos_x': pos_x_filtered,
                'pos_y': pos_y_filtered,
                'pos_z': pos_z_filtered,
                'L1': L1_filtered,
                'L2': L2_filtered,
                'L5': L5_filtered,
                'P1': P1_filtered,
                'P2': P2_filtered,
                'P5': P5_filtered,
                'cs_flag': cs_flag_filtered,
                'outlier_flag': outlier_flag_filtered,
                'satellite': satellite_filtered,
                'sta': sta_filtered,
                'hght': hght_filtered,
                'El': El_filtered,
                'Lon': Lon_filtered,
                'Lat': Lat_filtered,
                'obs_La': obs_La_filtered,
                'obs_Lb': obs_Lb_filtered,
                'obs_Lc': obs_Lc_filtered,
                'obs_Ca': obs_Ca_filtered,
                'obs_Cb': obs_Cb_filtered,
                'obs_Cc': obs_Cc_filtered
            }

            df = pd.DataFrame(data)

            # Converting relevant columns to float
            columns_to_convert = ['L1', 'L2', 'L5', 'P1', 'P2', 'P5']
            df[columns_to_convert] = df[columns_to_convert].astype(float)

            # Replace -999999.999 with NaN in relevant columns
            df.replace(-999999.999, np.nan, inplace=True)

            # Converting 'date' and 'time' columns to datetime and concatenating
            df['timestamp'] = pd.to_datetime(df['date'] + ' ' + df['time2'])

            # Converting lists to numpy arrays and ensuring float64 data types
            L1_array = np.nan_to_num(np.array(df['L1'].tolist(), dtype=np.float64), nan=-999999.999)
            L2_array = np.nan_to_num(np.array(df['L2'].tolist(), dtype=np.float64), nan=-999999.999)
            L5_array = np.nan_to_num(np.array(df['L5'].tolist(), dtype=np.float64), nan=-999999.999)

            P1_array = np.nan_to_num(np.array(df['P1'].tolist(), dtype=np.float64), nan=-999999.999)
            P2_array = np.nan_to_num(np.array(df['P2'].tolist(), dtype=np.float64), nan=-999999.999)
            P5_array = np.nan_to_num(np.array(df['P5'].tolist(), dtype=np.float64), nan=-999999.999)

            # Replace -999999.999 with NaN in P arrays
            L1_array[L1_array == -999999.999] = np.nan
            L2_array[L2_array == -999999.999] = np.nan
            L5_array[L5_array == -999999.999] = np.nan

            P1_array[P1_array == -999999.999] = np.nan
            P2_array[P2_array == -999999.999] = np.nan
            P5_array[P5_array == -999999.999] = np.nan

            # Find indices where df['outlier_flag'] is 'Y'
            outlier_indices = df.index[df['outlier_flag'] == 'Y'].tolist()

            # Convert these positions to NaN in all arrays
            for index in outlier_indices:
                L1_array[index] = np.nan
                L2_array[index] = np.nan
                L5_array[index] = np.nan
                P1_array[index] = np.nan
                P2_array[index] = np.nan
                P5_array[index] = np.nan

            L_GF = levelling_settings.geometry_free_combination_L(lambda1, lambda2, L1_array, L2_array)
            P_GF = levelling_settings.geometry_free_combination_C(P1_array, P2_array)

            L_GF15 = levelling_settings.geometry_free_combination_L(lambda1, lambda5, L1_array, L5_array)
            P_GF15 = levelling_settings.geometry_free_combination_C(P1_array, P5_array)

            L_GF25 = levelling_settings.geometry_free_combination_L(lambda2, lambda5, L2_array, L5_array)
            P_GF25 = levelling_settings.geometry_free_combination_C(P2_array, P5_array)

            DE = 3
            Thr = DE + 1
            arc_len = 15

            # Get indices where df['outlier_flag'] is equal to 'Y'
            indices_outliers = df.index[df['outlier_flag'] == 'Y']

            # Replace values in df['cs_flag'] with 'S' at the outlier indices
            df.loc[indices_outliers, 'cs_flag'] = 'S'

            # Assuming df['cs_flag'] is a pandas Series
            arcos = []  # List to store observation arcs
            arc_atual = []  # Temporary list to store current observation arc

            # Creating new column 'mini_flag' filled with 'N'
            df['mini_flag'] = 'N'

            # Identifying where 'outlier_flag' or 'cs_flag' are 'S' or 'Y' and replacing 'N' with 'Y' in 'mini_flag'
            mask = (df['outlier_flag'] == 'S') | (df['outlier_flag'] == 'Y') | (df['cs_flag'] == 'S') | (df['cs_flag'] == 'Y')
            df.loc[mask, 'mini_flag'] = 'Y'

            # Iterate over all elements in df['cs_flag']
            for idx, value in enumerate(df['mini_flag']):
                if value == 'Y':
                    # If current value is 'S', check if arc is not empty
                    # This avoids adding empty arcs in case of consecutive 'S'
                    if arc_atual:
                        arcos.append(arc_atual)
                        arc_atual = []  # Reset current arc list
                else:
                    # If value is not 'S', add index to current arc
                    arc_atual.append(idx)

            # Add the last arc if not empty
            if arc_atual:
                arcos.append(arc_atual)

            print()

            # Print info of each arc and classify them
            for i, arc in enumerate(arcos):
                start_index = arc[0]
                end_index = arc[-1]
                num_observations = len(arc)
                status = "Kept" if num_observations >= 15 else "Discarded"
                print(f"Arc {i + 1}: {df['timestamp'][start_index]} - {df['timestamp'][end_index]}, Start = {start_index}, End = {end_index}, "
                      f"Obs. = {num_observations}, Status = {status}")

            # Filter arcs that meet the length criterion
            arcos = [arc for arc in arcos if len(df['cs_flag'][arc[0]:arc[-1]+1]) >= 15]

            # Step 1: Create a copy of L_GF
            L_GF2 = np.copy(L_GF)

            # Step 2: Initialize a boolean array with True
            marcador_fora_arco = np.ones_like(L_GF, dtype=bool)

            # Step 3: Mark arc indices as False (not to convert to NaN)
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                marcador_fora_arco[start:end] = False

            # Step 4: Set values outside arcs to NaN
            L_GF2[marcador_fora_arco] = np.nan

            # Update L_GF to be the modified version
            L_GF = L_GF2

            # Reinitialize L_GF2 as a copy of L_GF
            L_GF2 = np.copy(L_GF)

            # Iterate over segments defined in 'arcos'
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                # Select current segment of L_GF
                segmento_data = L_GF[start:end]

                # Calculate first and third quartiles
                Q1 = np.nanpercentile(segmento_data, 25)
                Q3 = np.nanpercentile(segmento_data, 75)
                IQR = Q3 - Q1

                # Define limits for outliers
                lower_bound = Q1 - 8 * IQR
                upper_bound = Q3 + 8 * IQR

                # Find indices of outliers in the segment
                outlier_indices = np.where((segmento_data < lower_bound) | (segmento_data > upper_bound))

                # Replace outliers with NaN in L_GF2, not in segmento_data
                L_GF2[start:end][outlier_indices] = np.nan

            # Update L_GF to be the version without outliers
            L_GF = L_GF2

            # First, create a copy of P_GF to preserve original values
            P_GF_adjusted = np.array(P_GF, copy=True)  # Assuming P_GF is a NumPy array

            # Initialize all values as NaN
            P_GF_adjusted[:] = np.nan

            # Now, use a loop to assign non-NaN values within arcs
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                # Copy values from P_GF to P_GF_adjusted for the current arc
                P_GF_adjusted[start:end] = P_GF[start:end]

            # Now, P_GF_adjusted will have NaN outside the arcs and preserve original values inside
            L_GF_adjusted = list(L_GF)  # Initialize with original values

            limiar = 2  # Define the threshold

            # List to store subarcs
            subarcos = []

            # Plot original arcs in black (currently commented/placeholder)
            for i, arc in enumerate(arcos):
                L_GF_adjusted_arc = [L_GF_adjusted[i] for i in arc]
                P_GF_adjusted_arc = [P_GF[i] for i in arc]

            # Iterate over arcs
            for arc in arcos:
                # Filter L_GF_adjusted values for the current arc
                L_GF_adjusted_arc = [L_GF_adjusted[i] for i in arc]

                # Calculate difference between consecutive points
                valor = abs(np.diff(L_GF_adjusted_arc, prepend=np.nan))

                # Find indices where difference exceeds threshold
                outliers = np.where(valor > limiar)[0]

                # If there are outliers, split arc into subarcs
                if len(outliers) > 0:
                    subarco_indices = [arc[0]]  # Add arc start index

                    # Add outlier indices as subarc dividers
                    for outlier in outliers:
                        subarco_indices.append(arc[outlier])        # Divider
                        subarco_indices.append(arc[outlier] + 1)    # Next index after divider

                    subarco_indices.append(arc[-1])  # Add arc end index

                    # Create subarcs with calculated indices
                    for i in range(0, len(subarco_indices), 2):
                        subarcos.append(list(range(subarco_indices[i], subarco_indices[i+1] + 1)))
                else:
                    # If no outliers, keep the original arc
                    subarcos.append(arc)

            # Update arc list to use subarcs
            arcos = subarcos

            # Print information for each arc and classify them
            for i, arc in enumerate(arcos):
                start_index = arc[0]
                end_index = arc[-1]
                num_observations = len(arc)
                status = "Kept" if num_observations >= 15 else "Discarded"
                print(f"Arc {i + 1}: {df['timestamp'][start_index]} - {df['timestamp'][end_index]}, Start = {start_index}, End = {end_index}, "
                      f"Obs. = {num_observations}, Status = {status}")

            # Filter arcs that meet the minimum length criterion
            arcos = [arc for arc in arcos if len(df['cs_flag'][arc[0]:arc[-1]+1]) >= 15]

            # Step 1: Create a copy of L_GF
            L_GF2 = np.copy(L_GF)

            # Step 2: Initialize a boolean array with True
            marcador_fora_arco = np.ones_like(L_GF, dtype=bool)

            # Step 3: Mark the indices within arcs as False (do not convert to NaN)
            # Iterate over each valid arc
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                marcador_fora_arco[start:end] = False

            # Step 4: Set values outside arcs to NaN
            L_GF2[marcador_fora_arco] = np.nan

            # Update L_GF to be the modified version
            L_GF = L_GF2

            # Reinitialize L_GF2 as a copy of L_GF
            L_GF2 = np.copy(L_GF)

            # Iterate over the segments defined in 'arcos'
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                # Select current segment of L_GF
                segmento_data = L_GF[start:end]

                # Calculate first and third quartiles
                Q1 = np.nanpercentile(segmento_data, 25)
                Q3 = np.nanpercentile(segmento_data, 75)
                IQR = Q3 - Q1

                # Define outlier limits
                lower_bound = Q1 - 8 * IQR
                upper_bound = Q3 + 8 * IQR

                # Find indices of outliers within the segment
                outlier_indices = np.where((segmento_data < lower_bound) | (segmento_data > upper_bound))

                # Replace outliers with NaN in L_GF2
                L_GF2[start:end][outlier_indices] = np.nan

            # Update L_GF to be the version without outliers
            L_GF = L_GF2

            # First, create a copy of P_GF to preserve original values
            P_GF_adjusted = np.array(P_GF, copy=True)  # Assuming P_GF is a NumPy array

            # Initialize all values as NaN
            P_GF_adjusted[:] = np.nan

            # Assign values from P_GF to P_GF_adjusted inside the arcs
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                P_GF_adjusted[start:end] = P_GF[start:end]

            # Now, P_GF_adjusted has NaNs outside arcs and original values inside

            L_GF_adjusted = list(L_GF)  # Initialize with original values

            for arc in arcos:
                start = arc[0]
                end = arc[-1]
                # Calculate the difference and mean difference for the current arc
                diff = P_GF_adjusted[start:end] - L_GF_adjusted[start:end]
                mean_diff = np.nanmean(diff)

                # Adjust L_GF values inside this arc
                for j in range(start, end):
                    L_GF_adjusted[j] += mean_diff

            L_GF_adjusted_old = L_GF_adjusted

            # Define function to remove outliers based on the quartile method
            def remove_outliers_quartil(data):
                q1 = np.nanpercentile(data, 15)
                q3 = np.nanpercentile(data, 85)
                iqr = q3 - q1
                lower_bound = q1 - 1.3 * iqr
                upper_bound = q3 + 1.3 * iqr
                return [x for x in data if lower_bound <= x <= upper_bound]

            # Remove outliers individually for each arc
            for arc in arcos:
                start = arc[0]
                end = arc[-1]
                L_GF_adjusted[start:end+1] = remove_outliers_quartil(L_GF_adjusted[start:end+1])

            L_GF_adjusted2 = L_GF_adjusted
            P_GF_adjusted2 = P_GF_adjusted

            # Assume L_GF and P_GF are your original data series.
            # Initialize adjusted versions with NaNs
            L_GF_adjusted = np.full_like(L_GF, np.nan, dtype=np.float64)
            P_GF_adjusted = np.full_like(P_GF, np.nan, dtype=np.float64)

            polynomial_fits = []
            polynomial_fits2 = []
            arc_data = []
            arc_data2 = []

            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                # Update values in L_GF_adjusted and P_GF_adjusted only in defined arc intervals
                L_GF_adjusted[start:end] = L_GF[start:end]
                P_GF_adjusted[start:end] = P_GF[start:end]

                arc_values = L_GF[start:end]  # Use original data for fitting
                arc_values2 = P_GF[start:end]  # Use original data for fitting
                arc_timestamps = df['timestamp'][start:end]

                # Fit a polynomial of degree 3 to each set of values
                x_values = np.arange(len(arc_values))
                polynomial_fit = levelling_settings.fit_polynomial(x_values, arc_values, 3)

                x_values2 = np.arange(len(arc_values2))
                polynomial_fit2 = levelling_settings.fit_polynomial(x_values2, arc_values2, 3)

                # Store the fits and arc data
                polynomial_fits.append(polynomial_fit)
                polynomial_fits2.append(polynomial_fit2)
                arc_data.append(arc_values)
                arc_data2.append(arc_values2)

            L_GF_adjusted = list(L_GF_adjusted)  # Initialize with original values

            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                # Calculate difference and mean difference for the current arc
                diff = P_GF_adjusted[start:end] - L_GF_adjusted[start:end]
                mean_diff = np.nanmean(diff)

                # Adjust L_GF values inside this arc
                for i in range(start, end):
                    L_GF_adjusted[i] += mean_diff

            # Configure pandas to display all columns and rows
            pd.set_option('display.max_columns', None)  # Use None or set a number if preferred
            pd.set_option('display.max_rows', None)
            pd.set_option('display.max_colwidth', None)
            pd.set_option('display.width', None)

            L_GF_adjusted2 = L_GF_adjusted

            # Calculate quartiles for all values in L_GF_adjusted
            q1 = np.nanpercentile(L_GF_adjusted, 25)
            q3 = np.nanpercentile(L_GF_adjusted, 75)
            iqr = q3 - q1

            # Compute lower and upper bounds to identify outliers
            lower_bound = q1 - 2.5 * iqr
            upper_bound = q3 + 7.5 * iqr

            # Replace outliers with NaN across the entire dataset
            L_GF_adjusted = [x if lower_bound <= x <= upper_bound else np.nan for x in L_GF_adjusted]

            #plt.scatter(df['timestamp'], L_GF_adjusted, marker='o', s=20, color='blue', label='GF: L1-L2', zorder=0)

            # Escolhe cor e legenda com base na classe do satélite
            cor = 'navy' if sat_class == 'G' else 'red'
            label = 'GF: L1-L2 (GPS)' if sat_class == 'G' else 'GF: L1-L2 (GLONASS)'

            # Plota apenas UMA VEZ por classe para a legenda
            if satellite == satellites_to_plot[0]:
                plt.scatter(df['timestamp'], L_GF_adjusted, marker='o', s=20, color=cor, label=label, zorder=1)
            else:
                plt.scatter(df['timestamp'], L_GF_adjusted, marker='o', s=20, color=cor, zorder=1)


            # Ensure L_GF_adjusted and df have the same length
            if len(L_GF_adjusted) == len(df):
                df['LGF_combination'] = L_GF_adjusted
            else:
                print("Error: L_GF_adjusted and df do not have the same length!")

            # ---------------------- [L1-L5]

            # Step 1: Create a copy of L_GF15
            L_GF3 = np.copy(L_GF15)

            # Step 2: Initialize a boolean array with True
            marcador_fora_arco15 = np.ones_like(L_GF15, dtype=bool)

            # Step 3: Mark indices inside arcs as False (not to convert to NaN)
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                marcador_fora_arco15[start:end] = False

            # Step 4: Convert values outside arcs to NaN
            L_GF3[marcador_fora_arco15] = np.nan

            # Update L_GF15 to the modified version
            L_GF15 = L_GF3

            # Reinitialize L_GF3 as a copy of L_GF15
            L_GF3 = np.copy(L_GF15)

            # Iterate over segments defined by arcs
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                segmento_data = L_GF3[start:end]

                Q1 = np.nanpercentile(segmento_data, 25)
                Q3 = np.nanpercentile(segmento_data, 75)
                IQR = Q3 - Q1

                lower_bound = Q1 - 8 * IQR
                upper_bound = Q3 + 8 * IQR

                outlier_indices = np.where((segmento_data < lower_bound) | (segmento_data > upper_bound))
                L_GF3[start:end][outlier_indices] = np.nan

            L_GF15 = L_GF3

            # First, create a copy of P_GF15 to preserve original values
            P_GF_adjusted15 = np.array(P_GF15, copy=True)

            # Initialize all values as NaN
            P_GF_adjusted15[:] = np.nan

            # Assign values within arcs
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                P_GF_adjusted15[start:end] = P_GF15[start:end]

            # Initialize with original values
            L_GF_adjusted15 = list(L_GF15)

            limiar = 2  # Set appropriate threshold
            subarcos = []

            # Plot original arcs in black (placeholder loop)
            for i, arc in enumerate(arcos):
                L_GF_adjusted_arc15 = [L_GF_adjusted15[i] for i in arc]
                P_GF_adjusted_arc15 = [P_GF15[i] for i in arc]

            # Iterate over arcs
            for arc in arcos:
                L_GF_adjusted_arc15 = [L_GF_adjusted15[i] for i in arc]
                valor = abs(np.diff(L_GF_adjusted_arc15, prepend=np.nan))
                outliers = np.where(valor > limiar)[0]

                if len(outliers) > 0:
                    subarco_indices = [arc[0]]
                    for outlier in outliers:
                        subarco_indices.append(arc[outlier])
                        subarco_indices.append(arc[outlier] + 1)
                    subarco_indices.append(arc[-1])
                    for i in range(0, len(subarco_indices), 2):
                        subarcos.append(list(range(subarco_indices[i], subarco_indices[i+1] + 1)))
                else:
                    subarcos.append(arc)

            arcos = subarcos

            for i, arc in enumerate(arcos):
                start_index = arc[0]
                end_index = arc[-1]
                num_observations = len(arc)
                status = "Kept" if num_observations >= 15 else "Discarded"
                print(f"Arc {i + 1}: {df['timestamp'][start_index]} - {df['timestamp'][end_index]}, Start = {start_index}, End = {end_index}, "
                      f"Obs. = {num_observations}, Status = {status}")

            arcos = [arc for arc in arcos if len(df['cs_flag'][arc[0]:arc[-1]+1]) >= 15]

            # Step 1: Create a copy of L_GF
            L_GF3 = np.copy(L_GF15)

            # Step 2: Initialize a boolean array with True
            marcador_fora_arco15 = np.ones_like(L_GF15, dtype=bool)

            # Step 3: Mark indices inside arcs as False
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                marcador_fora_arco15[start:end] = False

            # Step 4: Convert values outside arcs to NaN
            L_GF3[marcador_fora_arco15] = np.nan

            L_GF15 = L_GF3
            L_GF3 = np.copy(L_GF15)

            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                segmento_data = L_GF15[start:end]

                Q1 = np.nanpercentile(segmento_data, 25)
                Q3 = np.nanpercentile(segmento_data, 75)
                IQR = Q3 - Q1

                lower_bound = Q1 - 8 * IQR
                upper_bound = Q3 + 8 * IQR

                outlier_indices = np.where((segmento_data < lower_bound) | (segmento_data > upper_bound))
                L_GF3[start:end][outlier_indices] = np.nan


            # Update L_GF to be the version without outliers
            L_GF15 = L_GF3

            # First, create a copy of P_GF to preserve the original values
            P_GF_adjusted15 = np.array(P_GF15, copy=True)  # Assuming P_GF is a list or NumPy array

            # Initialize all values as NaN
            P_GF_adjusted15[:] = np.nan

            # Assign non-NaN values inside arcs
            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                P_GF_adjusted15[start:end] = P_GF15[start:end]

            # Now P_GF_adjusted has NaN outside arcs and original values inside
            L_GF_adjusted15 = list(L_GF_adjusted15)  # Initialize with original values

            for arc in arcos:
                start = arc[0]
                end = arc[-1]
                diff15 = P_GF_adjusted15[start:end] - L_GF_adjusted15[start:end]
                mean_diff15 = np.nanmean(diff15)
                for j in range(start, end):
                    L_GF_adjusted15[j] += mean_diff15

            L_GF_adjusted_old15 = L_GF_adjusted15

            # Define function to remove outliers using the quartile method
            def remove_outliers_quartil(data):
                q1 = np.nanpercentile(data, 15)
                q3 = np.nanpercentile(data, 85)
                iqr = q3 - q1
                lower_bound = q1 - 1.3 * iqr
                upper_bound = q3 + 1.3 * iqr
                return [x for x in data if lower_bound <= x <= upper_bound]

            # Remove outliers individually for each arc
            for arc in arcos:
                start = arc[0]
                end = arc[-1]
                L_GF_adjusted15[start:end+1] = remove_outliers_quartil(L_GF_adjusted15[start:end+1])

            L_GF_adjusted3 = L_GF_adjusted15
            P_GF_adjusted3 = P_GF_adjusted15

            # Assume L_GF and P_GF are the original data series
            # Initialize adjusted versions with NaNs
            L_GF_adjusted15 = np.full_like(L_GF15, np.nan, dtype=np.float64)
            P_GF_adjusted15 = np.full_like(P_GF15, np.nan, dtype=np.float64)

            polynomial_fits = []
            polynomial_fits2 = []
            arc_data = []
            arc_data2 = []

            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                L_GF_adjusted15[start:end] = L_GF15[start:end]
                P_GF_adjusted15[start:end] = P_GF15[start:end]

                arc_values = L_GF15[start:end]  # Use original data for fitting
                arc_values2 = P_GF15[start:end]
                arc_timestamps = df['timestamp'][start:end]

                # Fit a 3rd-degree polynomial to each set
                x_values = np.arange(len(arc_values))
                polynomial_fit = levelling_settings.fit_polynomial(x_values, arc_values, 3)

                x_values2 = np.arange(len(arc_values2))
                polynomial_fit2 = levelling_settings.fit_polynomial(x_values2, arc_values2, 3)

                polynomial_fits.append(polynomial_fit)
                polynomial_fits2.append(polynomial_fit2)
                arc_data.append(arc_values)
                arc_data2.append(arc_values2)

            L_GF_adjusted15 = list(L_GF_adjusted15)  # Initialize with original values

            for i, arc in enumerate(arcos):
                start = arc[0]
                end = arc[-1]
                diff15 = P_GF_adjusted15[start:end] - L_GF_adjusted15[start:end]
                mean_diff15 = np.nanmean(diff15)
                for i in range(start, end):
                    L_GF_adjusted15[i] += mean_diff15

            # Configure pandas to display all columns and rows
            pd.set_option('display.max_columns', None)
            pd.set_option('display.max_rows', None)
            pd.set_option('display.max_colwidth', None)
            pd.set_option('display.width', None)

            L_GF_adjusted3 = L_GF_adjusted15

            # Calculate quartiles for all L_GF_adjusted values
            q1 = np.nanpercentile(L_GF_adjusted15, 25)
            q3 = np.nanpercentile(L_GF_adjusted15, 75)
            iqr = q3 - q1

            # Compute lower and upper bounds to identify outliers
            lower_bound = q1 - 2.5 * iqr
            upper_bound = q3 + 7.5 * iqr

            # Replace outliers with NaN
            L_GF_adjusted15 = [x if lower_bound <= x <= upper_bound else np.nan for x in L_GF_adjusted15]

            # Choose color and label based on satellite class
            cor = 'blue' if sat_class == 'G' else 'orange'
            label = 'GF: L1-L5 (GPS)' if sat_class == 'G' else 'GF: L2-L3 (GLONASS)'

            if satellite == satellites_to_plot[0]:
                plt.scatter(df['timestamp'], L_GF_adjusted15, marker='o', s=20, color=cor, label=label, zorder=2)
            else:
                plt.scatter(df['timestamp'], L_GF_adjusted15, marker='o', s=20, color=cor, zorder=2)

            # Ensure L_GF_adjusted has the same length as df
            if len(L_GF_adjusted15) == len(df):
                df['LGF_combination15'] = L_GF_adjusted15
            else:
                print("Error: L_GF_adjusted15 and df do not have the same length!")

            ########################################

            # Now df contains the LGF_combination column
            # List of desired columns
            colunas_desejadas = [
                'date', 'time2', 'mjd', 'pos_x', 'pos_y', 'pos_z',
                'LGF_combination', 'LGF_combination15', 'satellite',
                'sta', 'hght', 'El', 'Lon', 'Lat', 'obs_La', 'obs_Lb',
                'obs_Lc', 'obs_Ca', 'obs_Cb', 'obs_Cc'
            ]
            df_selecionado = df[colunas_desejadas]

            # Path and file name
            output_directory = os.path.join(str(ano), str(doy), estacao.upper())
            full_path = os.path.join(diretorio_principal)
            file_name = f"{estacao}_{satellite}_{doy}_{ano}.RNX3"
            fig_name = f"{estacao}_{doy}_{ano}.png"
            output_file_path = os.path.join(full_path, file_name)
            output_fig_path = os.path.join(full_path, fig_name)

            # Ensure directory exists
            os.makedirs(full_path, exist_ok=True)

            # Save selected DataFrame to tab-separated text file
            df_selecionado.to_csv(output_file_path, sep='\t', index=False, na_rep='-999999.999')

            print(f"Data exported to {output_file_path}.")

        plt.title(f"Station: {estacao.upper()}  |  Year: {ano}  |  DOY: {doy}", fontsize=16)
        plt.xlabel('Time (UT)', fontsize=16)
        plt.ylabel('Levelled Geometry-Free', fontsize=16)
        hours_fmt = mdates.DateFormatter('%H')
        plt.gca().xaxis.set_major_formatter(hours_fmt)
        minute_locator = mdates.MinuteLocator(interval=int1)
        plt.gca().xaxis.set_major_locator(minute_locator)
        plt.tick_params(axis='both', which='major', labelsize=14)
        plt.legend(bbox_to_anchor=(1.0, 1), loc='upper left')
        plt.grid(axis='both', linestyle='--', color='gray', linewidth=1)
        plt.tight_layout()
    if show_plot:
        plt.show()



from datetime import datetime
import matplotlib.pyplot as plt
import os
import sys
import numpy as np
import pandas as pd
import matplotlib.dates as mdates
from scipy.optimize import curve_fit
import itertools
import warnings
from numpy.polynomial import Polynomial
from pyOASIS import gnss_freqs
from pyOASIS import screening_settings
from pyOASIS import settings
import pyOASIS

def RNXScreening(destination_directory):
    # List of files in the .RNX1 directory
    filess = os.listdir(destination_directory)
    # Filter only the .RNX1 files
    files = [file_ for file_ in filess if file_.endswith("RNX1")]

    for file in files:
        f = os.path.join(destination_directory, file)
        # Get the file name
        g = os.path.basename(f)
        ano = g[13:17]
        doy = g[9:12]
        estacao = g[0:4]
        sat = g[5:8]

        # Variables and Parameters
        h1 = 0
        n_horas = 24 # hours
        int1 = 120 # minutes

        # Accessing the frequencies of the GPS system
        gps_freqs = gnss_freqs.FREQUENCY[gnss_freqs.GPS]
        f1 = gps_freqs[1]
        f2 = gps_freqs[2]
        f5 = gps_freqs[5]

        # Calculating the frequencies of the GLONASS system (glonass_channels.dat)
        # To obtain the GLONASS frequencies (code 'R'):
        file_name = os.path.join(pyOASIS.__path__[0], 'glonass_channels.dat')

        # Read the file into a DataFrame with predefined column names
        df_slots = pd.read_csv(file_name, sep=' ', header=None, names=['Slot', 'Channel'])
        glonass_frequencies = gnss_freqs.FREQUENCY[gnss_freqs.GLO]

        # List to store the data
        data = []

        # Iterate over each row in the DataFrame
        for index, row in df_slots.iterrows():
            satellite = row['Slot']
            k = row['Channel']
            row_data = [satellite]
            for channel, frequency in glonass_frequencies.items():
                if callable(frequency):  # Check if it is a lambda function
                    freq_value = frequency(k)
                else:
                    freq_value = frequency
                formatted_freq = f"{freq_value:.1f}"
                row_data.append(formatted_freq)
            # Add the row data to the data list
            data.append(row_data)

        # Convert the list of lists into a pandas DataFrame
        glo_freqs = pd.DataFrame(data, columns=['Satellite', 'fr1', 'fr2', 'fr3'])

        # Check the satellite class and adjust f1, f2, f5 accordingly
        if sat.startswith('G'):
            f1 = f1
            f2 = f2
            f5 = f5
        elif sat.startswith('R'):
            # Locate the row where 'Satellite' matches 'sat'
            sat_row = glo_freqs.loc[glo_freqs['Satellite'] == sat]
            if not sat_row.empty:
                f1 = float(sat_row['fr1'].values[0])
                f2 = float(sat_row['fr2'].values[0])
                f5 = float(sat_row['fr3'].values[0])
        else:
            f1 = f2 = f5 = None  # Or default values

        # Initialize lists to store values of each variable from all files
        date = []
        time = []
        mjd = []
        pos_x = []
        pos_y = []
        pos_z = []
        L1 = []
        L2 = []
        L5 = []
        P1 = []
        P2 = []
        P5 = []
        cs_flags = []
        satellites = []
        sta = []
        hght = []
        El = []
        Lon = []
        Lat = []
        obs_La = []
        obs_Lb = []
        obs_Lc = []
        obs_Ca = []
        obs_Cb = []
        obs_Cc = []

        caminho_arquivo = f

        with open(caminho_arquivo, 'r') as f:
            # Reading the file header
            header = f.readline().strip().split('\t')
            obs_La_header = header[6]
            obs_Lb_header = header[7]
            obs_Lc_header = header[8]
            obs_Ca_header = header[9]
            obs_Cb_header = header[10]
            obs_Cc_header = header[11]

            # Reading each data line from the file
            for linha in f:
                # Splitting the line into columns
                colunas = linha.strip().split('\t')  # Assuming tab-separated columns
                # Mapping each column to its corresponding header
                registro = {
                    'date': colunas[0],
                    'time': colunas[1],
                    'mjd': colunas[2],
                    'pos_x': colunas[3],
                    'pos_y': colunas[4],
                    'pos_z': colunas[5],
                    'L1': colunas[6],
                    'L2': colunas[7],
                    'L5': colunas[8],
                    'P1': colunas[9],
                    'P2': colunas[10],
                    'P5': colunas[11],
                    'cs_flags': colunas[12],
                    'satellite': colunas[13],
                    'sta': colunas[14],
                    'hght': colunas[15],
                    'El': colunas[16],
                    'Lon': colunas[17],
                    'Lat': colunas[18],
                    'obs_La': obs_La_header,
                    'obs_Lb': obs_Lb_header,
                    'obs_Lc': obs_Lc_header,
                    'obs_Ca': obs_Ca_header,
                    'obs_Cb': obs_Cb_header,
                    'obs_Cc': obs_Cc_header
                }

                # Appending each variable to its corresponding list
                date.append(registro['date'])
                time.append(registro['time'])
                mjd.append(registro['mjd'])
                pos_x.append(registro['pos_x'])
                pos_y.append(registro['pos_y'])
                pos_z.append(registro['pos_z'])
                L1.append(registro['L1'])
                L2.append(registro['L2'])
                L5.append(registro['L5'])
                P1.append(registro['P1'])
                P2.append(registro['P2'])
                P5.append(registro['P5'])
                cs_flags.append(registro['cs_flags'])
                satellites.append(registro['satellite'])
                sta.append(registro['sta'])
                hght.append(registro['hght'])
                El.append(registro['El'])
                Lon.append(registro['Lon'])
                Lat.append(registro['Lat'])
                obs_La.append(registro['obs_La'])
                obs_Lb.append(registro['obs_Lb'])
                obs_Lc.append(registro['obs_Lc'])
                obs_Ca.append(registro['obs_Ca'])
                obs_Cb.append(registro['obs_Cb'])
                obs_Cc.append(registro['obs_Cc'])

        # Initializing lists to store filtered values for each variable
        date_filtered = []
        time_filtered = []
        mjd_filtered = []
        pos_x_filtered = []
        pos_y_filtered = []
        pos_z_filtered = []
        L1_filtered = []
        L2_filtered = []
        L5_filtered = []
        P1_filtered = []
        P2_filtered = []
        P5_filtered = []
        cs_flags_filtered = []
        satellites_filtered = []
        sta_filtered = []
        hght_filtered = []
        El_filtered = []
        Lon_filtered = []
        Lat_filtered = []
        obs_La_filtered = []
        obs_Lb_filtered = []
        obs_Lc_filtered = []
        obs_Ca_filtered = []
        obs_Cb_filtered = []
        obs_Cc_filtered = []

        satellite = sat
        print(f"Processing: {satellite}")
        indices = np.where(np.array(satellites) == satellite)[0]

        date_filtered = []
        time_filtered = []
        mjd_filtered = []
        pos_x_filtered = []
        pos_y_filtered = []
        pos_z_filtered = []
        L1_filtered = []
        L2_filtered = []
        L5_filtered = []
        P1_filtered = []
        P2_filtered = []
        P5_filtered = []
        cs_flags_filtered = []
        satellites_filtered = []
        sta_filtered = []
        hght_filtered = []
        El_filtered = []
        Lon_filtered = []
        Lat_filtered = []
        obs_La_filtered = []
        obs_Lb_filtered = []
        obs_Lc_filtered = []
        obs_Ca_filtered = []
        obs_Cb_filtered = []
        obs_Cc_filtered = []

        # Filtering the values for the selected satellite
        for idx in indices:
            date_filtered.append(date[idx])
            time_filtered.append(time[idx])
            mjd_filtered.append(mjd[idx])
            pos_x_filtered.append(pos_x[idx])
            pos_y_filtered.append(pos_y[idx])
            pos_z_filtered.append(pos_z[idx])
            L1_filtered.append(L1[idx])
            L2_filtered.append(L2[idx])
            L5_filtered.append(L5[idx])
            P1_filtered.append(P1[idx])
            P2_filtered.append(P2[idx])
            P5_filtered.append(P5[idx])
            cs_flags_filtered.append(cs_flags[idx])
            satellites_filtered.append(satellites[idx])
            sta_filtered.append(sta[idx])
            hght_filtered.append(hght[idx])
            El_filtered.append(El[idx])
            Lon_filtered.append(Lon[idx])
            Lat_filtered.append(Lat[idx])
            obs_La_filtered.append(obs_La[idx])
            obs_Lb_filtered.append(obs_Lb[idx])
            obs_Lc_filtered.append(obs_Lc[idx])
            obs_Ca_filtered.append(obs_Ca[idx])
            obs_Cb_filtered.append(obs_Cb[idx])
            obs_Cc_filtered.append(obs_Cc[idx])

        # Building a DataFrame with the filtered data for the selected satellite
        data = {
            'date': date_filtered,
            'time2': time_filtered,
            'mjd': mjd_filtered,
            'pos_x': pos_x_filtered,
            'pos_y': pos_y_filtered,
            'pos_z': pos_z_filtered,
            'L1': L1_filtered,
            'L2': L2_filtered,
            'L5': L5_filtered,
            'P1': P1_filtered,
            'P2': P2_filtered,
            'P5': P2_filtered,
            'cs_flag': cs_flags_filtered,
            'satellite': satellites_filtered,
            'sta': sta_filtered,
            'hght': hght_filtered,
            'El': El_filtered,
            'Lon': Lon_filtered,
            'Lat': Lat_filtered,
            'obs_La': obs_La_filtered,
            'obs_Lb': obs_Lb_filtered,
            'obs_Lc': obs_Lc_filtered,
            'obs_Ca': obs_Ca_filtered,
            'obs_Cb': obs_Cb_filtered,
            'obs_Cc': obs_Cc_filtered
        }

        df = pd.DataFrame(data)

        # Convert relevant columns to float
        columns_to_convert = ['L1', 'L2', 'L5', 'P1', 'P2', 'P5']
        df[columns_to_convert] = df[columns_to_convert].astype(float)

        # Replace -999999.999 with NaN in relevant columns
        df.replace(-999999.999, np.nan, inplace=True)

        # Convert 'date' and 'time' columns to datetime and combine them
        df['timestamp'] = pd.to_datetime(df['date'] + ' ' + df['time2'])

        # Extract only the time from 'timestamp' and store in a new column 'time'
        df['time'] = df['timestamp'].dt.time

        # Convert lists to numpy arrays and ensure they are float64
        L1_array = np.nan_to_num(np.array(df['L1'].tolist(), dtype=np.float64), nan=-999999.999)
        L2_array = np.nan_to_num(np.array(df['L2'].tolist(), dtype=np.float64), nan=-999999.999)
        L5_array = np.nan_to_num(np.array(df['L5'].tolist(), dtype=np.float64), nan=-999999.999)

        P1_array = np.nan_to_num(np.array(df['P1'].tolist(), dtype=np.float64), nan=-999999.999)
        P2_array = np.nan_to_num(np.array(df['P2'].tolist(), dtype=np.float64), nan=-999999.999)
        P5_array = np.nan_to_num(np.array(df['P5'].tolist(), dtype=np.float64), nan=-999999.999)

        # Replace -999999.999 with NaN in the arrays
        L1_array[L1_array == -999999.999] = np.nan
        L2_array[L2_array == -999999.999] = np.nan
        L5_array[L5_array == -999999.999] = np.nan

        P1_array[P1_array == -999999.999] = np.nan
        P2_array[P2_array == -999999.999] = np.nan
        P5_array[P5_array == -999999.999] = np.nan

        # Compute the Melbourne-Wübbena combination for the current satellite
        MW_combination = screening_settings.melbourne_wubbena_combination(f1, f2, L1_array, L2_array, P1_array, P2_array)
        MW_combination2 = screening_settings.melbourne_wubbena_combination(f1, f5, L1_array, L5_array, P1_array, P5_array)

        IFL_combination = screening_settings.iono_free_phase_combination(f1, f2, L1_array, L2_array)
        IFP_combination = screening_settings.iono_free_range_combination(f1, f2, P1_array, P2_array)

        # Add MW, ionosphere-free phase, and ionosphere-free code combinations to the DataFrame
        df['MW'] = MW_combination
        df['MW2'] = MW_combination2

        # Assuming df['cs_flag'] is a pandas Series
        arcos = []  # List to store observation arcs
        arc_atual = []  # Temporary list to store current observation arc

        # Iterate over all elements in df['cs_flag']
        for idx, value in enumerate(df['cs_flag']):
            if value == 'S':
                # If the current value is 'S', check if the current arc is not empty
                # This avoids appending empty arcs in case of consecutive 'S'
                if arc_atual:
                    arcos.append(arc_atual)
                    arc_atual = []  # Reseta a lista do arco atual
            else:
                # If the value is not 'S', add index to the current arc
                arc_atual.append(idx)

        # Append the last arc if it's not empty
        if arc_atual:
            arcos.append(arc_atual)

        print()

        # Print information about each arc and classify them
        for i, arc in enumerate(arcos):
            start_index = arc[0]
            end_index = arc[-1]
            num_observations = len(arc)
            status = "Kept" if num_observations >= 15 else "Discarded"
            print(f"Arc {i + 1}: {df['timestamp'][start_index]} - {df['timestamp'][end_index]}, Start = {start_index}, End = {end_index}, "
                f"Obs. = {num_observations}, Status = {status}")

        # Separate MW_combination data for each arc
        arc_data = []
        arc_idx = []
        polynomial_fits = []

        print()

        for i, arc in enumerate(arcos):
            start = arc[0]
            end = arc[-1]
            arc_values = MW_combination[start:end+1]
            arc_timestamps = df['timestamp'][start:end+1]

            if len(arc_values) < 15:
                continue

            # Fit a 3rd-degree polynomial
            x_values = np.arange(len(arc_values))
            polynomial_fit = screening_settings.fit_polynomial(x_values, arc_values, 3)

            # Store arc values and the polynomial fit
            arc_data.append(arc_values)
            arc_idx.append(arc_timestamps)
            polynomial_fits.append(polynomial_fit)

            # Print arc information
            num_observations = len(arc_values)
            num_points_fit = len(polynomial_fit)

            print(f"Arc {i + 1}: Start index = {start}, End index = {end}, "
                f"Number of observations = {num_observations}, Number of fit points = {num_points_fit}")

        # Filter arcs that meet the minimum length criterion
        arcos_validos = [arc for arc in arcos if len(MW_combination[arc[0]:arc[-1]+1]) >= 15]

        # If there's only one valid arc, duplicate it to ensure at least two
        if len(arcos_validos) == 1:
            arcos_validos.append(arcos_validos[0])

        # Set the maximum number of columns per row
        max_colunas_por_linha = 2

        # Calculate the number of rows and columns needed
        num_arcos_validos = len(arcos_validos)

        num_linhas = (num_arcos_validos - 1) // max_colunas_por_linha + 1
        num_colunas = min(num_arcos_validos, max_colunas_por_linha)

        all_all = []
        all_index = []

        # Iterate over each valid arc and plot the data
        for i, (arc) in enumerate(arcos_validos, start=1):
            start = arc[0]
            end = arc[-1]
            arc_data = df.iloc[arc]
            time = df.index[arc]

            arc_values = MW_combination[int(start):int(end)+1]
            arc_timestamps = df['timestamp'][start:end+1]
            arc_values2 = arc_values

            # Calculate elapsed time in seconds from the first timestamp in the arc
            x = (arc_timestamps - arc_timestamps.iloc[0]).dt.total_seconds()

            y_rescaled = screening_settings.rescale_data(arc_values)
            delta_y = np.diff(y_rescaled, prepend=np.nan)

            # Fit a polynomial only to valid values (excluding NaNs)
            p = Polynomial.fit(x[1:], delta_y[1:], 3)

            delta_y_fit = p(x)  # Polynomial-fitted values
            residuals = delta_y - delta_y_fit  # Compute residuals

            mini_arcos = []  # List to store mini-arcs
            mini_arcos_mantidos = []
            mini_arc_atual = []  # Temporary list to store the current mini-arc
            signo_anterior = None

            # Iterate over all residual values
            for idx, value in enumerate(residuals):
                if signo_anterior is None:  # If it's the first value, initialize the sign
                    signo_anterior = np.sign(value)

                # Check if the sign has changed
                if np.sign(value) != signo_anterior:
                    # If the current mini-arc is not empty, add it to the list
                    if mini_arc_atual:
                        mini_arcos.append(mini_arc_atual)
                    mini_arc_atual = []  # Start a new mini-arc

                # Add the index to the current mini-arc
                mini_arc_atual.append(idx)

                # Update the previous sign
                signo_anterior = np.sign(value)

            # List to store mini-arcs that meet the minimum observation requirement
            mini_arcos_mantidos = []

            # Flag to control early termination of the loop
            should_break = False

            print()
            print("Looking for mini cycle-slips in L1-L2 pair:")
            print()

            # Outer loop
            for mini_i, mini_arc in enumerate(mini_arcos):
                mini_start_index = mini_arc[0]
                mini_end_index = mini_arc[-1]
                num_observations = len(mini_arc)
                status = "Kept" if num_observations >= 4 else "Discarded"
                print(f"Mini-arc {mini_i + 1}: Start = {mini_start_index}, End = {mini_end_index}, Obs. = {num_observations}, Status = {status}")

                # If the number of observations is less than 4, set the flag and continue
                if num_observations <= 4:
                    should_break = True
                    continue

                # If the mini-arc has at least 4 observations, keep it
                mini_arcos_mantidos.append(mini_arc)

            # If the flag indicates we should exit the loops, break from the outer loop as well
            if should_break:
                continue

            # Print the kept mini-arcs
            print("Kept mini-arcs:")
            for i, mini_arc in enumerate(mini_arcos_mantidos):
                print(f"Mini-arc {i + 1}: {mini_arc}")

            # Quartile and IQR calculation for identifying outliers
            Q1 = np.nanpercentile(residuals, 15)
            Q3 = np.nanpercentile(residuals, 85)
            IQR = Q3 - Q1

            threshold2 = 2

            outlier_mask = (residuals < Q1 - threshold2 * IQR) | (residuals > Q3 + threshold2 * IQR)
            high_residuals_mask = residuals > 1  # Mask for high residuals
            other_residuals_mask = ~(outlier_mask | high_residuals_mask)  # Mask for remaining residuals

            # Fit a third-degree polynomial
            x_values = np.arange(len(arc_values))
            polynomial_fit = screening_settings.fit_polynomial(x_values, arc_values, 3)

            # Get the corresponding arc number
            num_arc_valido = arcos.index(arc) + 1

            all_out = []  # Initialize 'all_out' list before the loop
            all__all = []
            all_indices = []    # List to store indices of high residuals and outliers
            todos_indices = []  # List to store all indices

            for i, (mini_start, mini_end) in enumerate(mini_arcos_mantidos):
                mini_residuals = residuals[mini_start:mini_end]
                mini_time = time[mini_start:mini_end]
                try:
                    mini_fit = screening_settings.fit_polynomial(mini_time, mini_residuals, 3)
                    new_mini_residuals = abs(mini_residuals-mini_fit)

                    # Quartile and IQR calculation for identifying outliers
                    mini_Q1 = np.nanpercentile(new_mini_residuals, 15)
                    mini_Q3 = np.nanpercentile(new_mini_residuals, 85)
                    mini_IQR = mini_Q3 - mini_Q1

                    mini_threshold2 = 1.3

                    mini_outlier_mask = (new_mini_residuals < mini_Q1 - mini_threshold2 * mini_IQR) | (new_mini_residuals > mini_Q3 + mini_threshold2 * mini_IQR)
                    mini_high_residuals_mask = new_mini_residuals > 0.0002  # Mask for high residuals
                    mini_other_residuals_mask = ~(mini_outlier_mask | mini_high_residuals_mask)  # Mask for remaining residuals

                    micro_residuals = abs(np.diff(new_mini_residuals, prepend=np.nan))

                    # Compute quartiles
                    q1_micro = np.nanpercentile(micro_residuals, 15)
                    q3_micro = np.nanpercentile(micro_residuals, 85)

                    # Compute interquartile range
                    iqr_micro = q3_micro - q1_micro

                    micro_threshold2 = 1.3

                    # Define the bounds for identifying outliers
                    lower_bound_micro = q1_micro - micro_threshold2 * iqr_micro
                    upper_bound_micro = q3_micro + micro_threshold2 * iqr_micro

                    # Identify outliers
                    outliers_micro = (micro_residuals < lower_bound_micro) | (micro_residuals > upper_bound_micro)

                    micro_high_residuals_mask = micro_residuals > 0.00002  # Mask for high residuals

                    # Convert local indices to global indices – mini
                    global_high_residuals_indices = np.where(mini_high_residuals_mask)[0] + mini_start
                    global_outlier_indices = np.where(mini_outlier_mask)[0] + mini_start

                    # Convert local indices to global indices – micro
                    global_high_residuals_indices_micro = np.where(micro_high_residuals_mask)[0] + mini_start
                    global_outlier_indices_micro = np.where(outliers_micro)[0] + mini_start

                    # Lists to store the indices of high residuals and outliers
                    indices_residuos_altos_mini = start + global_high_residuals_indices
                    indices_outliers_mini = start + global_outlier_indices
                    indices_residuos_altos_micro = start + global_high_residuals_indices_micro
                    indices_outliers_micro = start + global_outlier_indices_micro

                    # Add the indices to the general list
                    todos_indices.append({
                        'residuos_altos_mini': indices_residuos_altos_mini,
                        'outliers_mini': indices_outliers_mini,
                        'residuos_altos_micro': indices_residuos_altos_micro,
                        'outliers_micro': indices_outliers_micro
                    })

                except:
                    pass  # Ignore and continue the loop

            # Create a copy of the all_indexes list before extending it to all_index_vertical
            todos_indices_copy = todos_indices.copy()

            # Access all indices outside the loop
            for indices in todos_indices_copy:
                print(f"Indices for iteration {i}:")
                print("High Residuals for Mini-Arc:", indices['residuos_altos_mini'])
                print("Outliers for Mini-Arc:", indices['outliers_mini'])
                print("High Residuals for Micro-Arc:", indices['residuos_altos_micro'])
                print("Outliers for Micro-Arc:", indices['outliers_micro'])
                print()

            # Concatenate all indices into a vertical list
            todos_indices_vertical = []
            for indices in todos_indices_copy:
                todos_indices_vertical.extend(indices['residuos_altos_mini'])
                todos_indices_vertical.extend(indices['outliers_mini'])
                todos_indices_vertical.extend(indices['residuos_altos_micro'])
                todos_indices_vertical.extend(indices['outliers_micro'])

            from collections import OrderedDict

            # Convert all_indices_vertical to a set to remove duplicates, then back to a list maintaining original order
            todos_indices_vertical = list(OrderedDict.fromkeys(todos_indices_vertical))
            all_index.extend(todos_indices_vertical)

            # Remove empty entries from the list
            all_index = list(filter(None, all_index))

        print()
        print(satellite)

        df['outlier_flag'] = 'N'

        # Replace 'N' with 'Y' at the specified indices
        df.loc[all_index, 'outlier_flag'] = 'Y'

        # ---- [L1 - L5]
        all_index15 = []

        # Iterate over each valid arc and plot the data
        #for i, (arc) in enumerate(zip(arcos_validos), start=1):
        for i, arc in enumerate(arcos_validos, start=1):
            start = arc[0]
            end = arc[-1]
            arc_data = df.iloc[arc]
            time = df.index[arc]

            arc_values = MW_combination2[start:end+1]
            arc_timestamps = df['timestamp'][start:end+1]
            arc_values2 = arc_values

            # Compute elapsed time in seconds from the first timestamp in the arc
            x = (arc_timestamps - arc_timestamps.iloc[0]).dt.total_seconds()

            y_rescaled = screening_settings.rescale_data(arc_values)
            delta_y = np.diff(y_rescaled, prepend=np.nan)

            # Fit a polynomial only to valid values (excluding NaN)
            p = Polynomial.fit(x[1:], delta_y[1:], 3)

            delta_y_fit = p(x)  # Fitted values
            residuals = delta_y - delta_y_fit  # Compute residuals

            mini_arcos = []  # List to store observation mini-arcs
            mini_arcos_mantidos = []
            mini_arc_atual = []  # Temporary list for current mini-arc
            signo_anterior = None

            # Iterate through all residual values
            for idx, value in enumerate(residuals):
                if signo_anterior is None:  # First value, initialize previous_sign
                    signo_anterior = np.sign(value)

                # Check if the sign changed
                if np.sign(value) != signo_anterior:
                    if mini_arc_atual:
                        mini_arcos.append(mini_arc_atual)
                    mini_arc_atual = []
                mini_arc_atual.append(idx)
                signo_anterior = np.sign(value)

            # List to store mini-arcs that pass the minimum criteria
            mini_arcos_mantidos = []
            should_break = False

            print()
            print("Looking for mini cycle-slips in L1-L5 pair:")
            print()

            # Outer loop
            for mini_i, mini_arc in enumerate(mini_arcos):
                mini_start_index = mini_arc[0]
                mini_end_index = mini_arc[-1]
                num_observations = len(mini_arc)
                status = "Kept" if num_observations >= 4 else "Discarded"
                print(f"Mini-arc {mini_i + 1}: Start = {mini_start_index}, End = {mini_end_index}, Obs. = {num_observations}, Status = {status}")

                if num_observations <= 4:
                    should_break = True
                    continue

                mini_arcos_mantidos.append(mini_arc)

            if should_break:
                continue

            print("Kept mini-arcs:")
            for i, mini_arc in enumerate(mini_arcos_mantidos):
                print(f"Mini-arc {i + 1}: {mini_arc}")

            # Compute quartiles and IQR to identify outliers
            Q1 = np.nanpercentile(residuals, 15)
            Q3 = np.nanpercentile(residuals, 85)
            IQR = Q3 - Q1

            threshold2 = 2

            outlier_mask = (residuals < Q1 - threshold2 * IQR) | (residuals > Q3 + threshold2 * IQR)
            high_residuals_mask = residuals > 1  # Máscara para resíduos altos
            other_residuals_mask = ~(outlier_mask | high_residuals_mask)  # Máscara para os demais resíduos

            # Fit a 3rd-degree polynomial
            x_values = np.arange(len(arc_values))
            polynomial_fit = screening_settings.fit_polynomial(x_values, arc_values, 3)

            #ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
            num_arc_valido = arcos.index(arc) + 1

            all_out = []
            all__all = []
            all_indices = []
            todos_indices = []

            for i, (mini_start, mini_end) in enumerate(mini_arcos_mantidos):
                mini_residuals = residuals[mini_start:mini_end]
                mini_time = time[mini_start:mini_end]
                try:
                    mini_fit = screening_settings.fit_polynomial(mini_time, mini_residuals, 3)
                    new_mini_residuals = abs(mini_residuals-mini_fit)

                    # Quartile and IQR for outlier detection
                    mini_Q1 = np.nanpercentile(new_mini_residuals, 15)
                    mini_Q3 = np.nanpercentile(new_mini_residuals, 85)
                    mini_IQR = mini_Q3 - mini_Q1

                    mini_threshold2 = 1.3

                    mini_outlier_mask = (new_mini_residuals < mini_Q1 - mini_threshold2 * mini_IQR) | (new_mini_residuals > mini_Q3 + mini_threshold2 * mini_IQR)
                    mini_high_residuals_mask = new_mini_residuals > 0.0002  # Mask for high residuals
                    mini_other_residuals_mask = ~(mini_outlier_mask | mini_high_residuals_mask)  # Mask for the remaining residuals

                    micro_residuals = abs(np.diff(new_mini_residuals, prepend=np.nan))

                    # Calculate Q1 and Q3
                    q1_micro = np.nanpercentile(micro_residuals, 15)
                    q3_micro = np.nanpercentile(micro_residuals, 85)

                    # Calculate IQR
                    iqr_micro = q3_micro - q1_micro
                    micro_threshold2 = 1.3

                    lower_bound_micro = q1_micro - micro_threshold2 * iqr_micro
                    upper_bound_micro = q3_micro + micro_threshold2 * iqr_micro

                    # Find outliers
                    outliers_micro = (micro_residuals < lower_bound_micro) | (micro_residuals > upper_bound_micro)
                    micro_high_residuals_mask = micro_residuals > 0.00002

                    # Convert local to global indices – mini
                    global_high_residuals_indices = np.where(mini_high_residuals_mask)[0] + mini_start
                    global_outlier_indices = np.where(mini_outlier_mask)[0] + mini_start

                    # Convert local to global indices – micro
                    global_high_residuals_indices_micro = np.where(micro_high_residuals_mask)[0] + mini_start
                    global_outlier_indices_micro = np.where(outliers_micro)[0] + mini_start

                    # Lists to store indices of high residuals and outliers
                    indices_residuos_altos_mini = start + global_high_residuals_indices
                    indices_outliers_mini = start + global_outlier_indices
                    indices_residuos_altos_micro = start + global_high_residuals_indices_micro
                    indices_outliers_micro = start + global_outlier_indices_micro

                    indices_residuos_altos_mini = start + global_high_residuals_indices
                    indices_outliers_mini = start + global_outlier_indices
                    indices_residuos_altos_micro = start + global_high_residuals_indices_micro
                    indices_outliers_micro = start + global_outlier_indices_micro

                    todos_indices.append({
                        'residuos_altos_mini': indices_residuos_altos_mini,
                        'outliers_mini': indices_outliers_mini,
                        'residuos_altos_micro': indices_residuos_altos_micro,
                        'outliers_micro': indices_outliers_micro
                    })

                except:
                    pass  # Ignore and continue

            # Create a copy of the todos_indices list before extending todos_indices_vertical
            todos_indices_copy = todos_indices.copy()

            # Now you can access all indices outside the loop
            for indices in todos_indices_copy:
                print(f"Indices for iteration {i}:")
                print("High Residuals for Mini-Arc:", indices['residuos_altos_mini'])
                print("Outliers for Mini-Arc:", indices['outliers_mini'])
                print("High Residuals for Micro-Arc:", indices['residuos_altos_micro'])
                print("Outliers for Micro-Arc:", indices['outliers_micro'])
                print()

            # Concatenate all indices into a vertical list
            todos_indices_vertical = []
            for indices in todos_indices_copy:
                todos_indices_vertical.extend(indices['residuos_altos_mini'])
                todos_indices_vertical.extend(indices['outliers_mini'])
                todos_indices_vertical.extend(indices['residuos_altos_micro'])
                todos_indices_vertical.extend(indices['outliers_micro'])

            from collections import OrderedDict

            # Convert todos_indices_vertical into a set to remove duplicates and then back to a list preserving original order
            todos_indices_vertical = list(OrderedDict.fromkeys(todos_indices_vertical))

            all_index15.extend(todos_indices_vertical)

            # Remove empty entries from the list
            all_index15 = list(filter(None, all_index15))

        print()
        print(satellite)

        # Replace 'N' with 'Y' at the specified indices
        df.loc[all_index15, 'outlier_flag'] = 'Y'

        # Output directory and desired path
        output_directory = os.path.join(str(ano), str(doy), estacao.upper())

        # Full output directory path within the destination directoryo
        full_path = os.path.join(destination_directory)

        # Ensure the directory exists or create it if not
        os.makedirs(full_path, exist_ok=True)

        # Define the file name
        file_name = f"{estacao}_{satellite}_{doy}_{ano}.RNX2"

        # Full path of the output file
        output_file_path = os.path.join(full_path, file_name)

        # Select only the desired columns
        colunas_desejadas = ['date', 'time', 'mjd', 'pos_x', 'pos_y', 'pos_z', 'L1', 'L2', 'L5', 'P1', 'P2', 'P5', 'cs_flag', 'outlier_flag', 'satellite', 'sta', 'hght', 'El', 'Lon', 'Lat', 'obs_La', 'obs_Lb', 'obs_Lc', 'obs_Ca', 'obs_Cb', 'obs_Cc']

        df_selecionado = df[colunas_desejadas]

        # Replace NaN with -999999.999
        df_selecionado = df_selecionado.fillna(-999999.999)

        # Save the selected DataFrame to a tab-delimited text file
        df_selecionado.to_csv(output_file_path, sep='\t', index=False, na_rep='-999999.999')

        # Read the file to verify it is correct
        with open(output_file_path, 'r') as f:
            file_content = f.read()

        print(f"Data exported to {output_file_path}.")



import os
import numpy as np
import sys
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from astropy.time import Time
from scipy.constants import speed_of_light
from pyOASIS import gnss_freqs

# Function definitions
def geometry_free_combination_L(lambda1, lambda2, L1, L2):
    """
    Computes the geometry-free combination for carrier phase measurements.

    Parameters:
    - lambda1, lambda2: Wavelengths for the two frequencies.
    - L1, L2: Phase observations for the two frequencies.

    Returns:
    - L_GF: Result of the geometry-free combination for carrier phase.
    """
    L_GF = (lambda1*L1)-(lambda2*L2)
    # L_GF = L1 - L2
    return L_GF

def geometry_free_combination_C(P1, P2):
    """
    Computes the geometry-free combination for pseudorange code measurements.

    Parameters:
    - P1, P2: Code observations for the two frequencies.

    Returns:
    - P_GF: Result of the geometry-free combination for code.
    """
    P_GF = P2-P1
    return P_GF

# Function to fit a low-degree polynomial to data
def fit_polynomial(x, y, degree):
    coeffs = np.polyfit(x, y, degree)
    return np.polyval(coeffs, x)

def detect_outliers(arc_data, polynomial_fits, arc_idx, threshold_factor):
    outliers = []
    for i, (arc_values, fit) in enumerate(zip(arc_data, polynomial_fits), start=1):
        # Compute the residuals
        residuals = np.abs(arc_values - fit)
        # Compute the median of residuals
        mean_residuals = np.median(residuals)
        # Define the threshold for outlier detection
        threshold = threshold_factor * mean_residuals
        # Identify the values that are outliers
        outlier_indices = np.where(np.abs(residuals) > threshold)[0]
        # Retrieve the current arc limits
        arc_start, arc_end = arc_idx[i - 1][0], arc_idx[i - 1][-1] + 1
        # Transform arc-local outlier indices into global indices
        real_outlier_indices = outlier_indices + arc_start
        # Add outliers to the list
        outliers.extend([(i, arc_values[idx], idx, real_idx) for idx, real_idx in zip(outlier_indices, real_outlier_indices)])
    return outliers



from pyproj import Transformer, CRS
import numpy as np

def convert_coords(obs_x, obs_y, obs_z, to_radians = True):
    """
    Converts cartesian to geodesic coordinates.
    Just now, just for the receiver positions.
    The parameters must be in meters [m]
    >>> obs_x, obs_y, obs_z = 5043729.726, -3753105.556, -1072967.067
    >>> print(convert_coords(obs_x, obs_y, obs_z))
    ...-36.6534, -9.7492, 266.23012
    """

    crs_from = CRS(proj = 'geocent', ellps = 'WGS84', datum = 'WGS84')

    crs_to = CRS(proj = 'latlong', ellps = 'WGS84', datum = 'WGS84')

    transformer = Transformer.from_crs(crs_from, crs_to)

    lon, lat, alt = transformer.transform(xx = obs_x,
                                          yy = obs_y,
                                          zz = obs_z,
                                          radians = False)

    if to_radians:
        lon = np.radians(lon + 360)
        lat = np.radians(lat)

    return lon, lat, alt


Re  = 6371.0
hm = 250.0
dtor = 0.0174533


class IonosphericPiercingPoint(object):

    def __init__(self,
                 sat_x, sat_y, sat_z,
                 obs_x, obs_y, obs_z):



        # converts meters to kilometers
        obs_x /= 1000
        obs_y /= 1000
        obs_z /= 1000

        # Compute the relative positions
        self.dx = (sat_x - obs_x)
        self.dy = (sat_y - obs_y)
        self.dz = (sat_z - obs_z)

        self.dxdy = (sat_x * obs_y - sat_y * obs_x)
        self.dxdz = (sat_x * obs_z - sat_z * obs_x)

    def positions(self, height = "top"):

        """Positions of ionospheric piercing point for
        differents altitudes"""

        if height == "top":
            h = 6870.0
        else:
            h = 6620.0

        arg_1 = pow((self.dy * self.dxdy + self.dz * self.dxdz), 2)
        arg_2 = pow(self.dx, 2) + pow(self.dy, 2) + pow(self.dz, 2)
        arg_3 = pow(self.dxdy, 2) + pow(self.dxdz, 2) - pow(self.dx, 2) * pow(h, 2)
        arg_4 = pow(self.dy, 2) + pow(self.dz, 2) + pow(self.dx, 2)

        factor = np.sqrt(arg_1 - arg_2 * arg_3)

        sub_ion_x = (-1 * (self.dy * self.dxdy +
                           self.dz * self.dxdz) -
                     factor) / arg_4

        sub_ion_y = (self.dy * sub_ion_x + self.dxdy) / self.dx
        sub_ion_z = (self.dz * sub_ion_x + self.dxdz) / self.dx

        return sub_ion_x, sub_ion_y, sub_ion_z

    def relative_directions(self, lat, lon):

        """
        Relative directions in function of latitude and
        longitude of receiver (in degrees)
        """

        meridional = (-np.cos(lon) * np.sin(lat) * self.dx -
                      np.sin(lon) * np.sin(lat) * self.dy +
                      np.cos(lat) * self.dz)

        zonal = - np.sin(lon) * self.dx + np.cos(lon) * self.dy

        vertical = (np.cos(lon) * np.cos(lat) * self.dx +
                    np.sin(lon) * np.cos(lat) * self.dy +
                    np.sin(lat) * self.dz)

        return meridional, zonal, vertical

    def zenital_angle(self, lat, lon):

        """Zenital angle between satellite and receiver"""

        meridional, zonal, vertical = self.relative_directions(lat, lon)

        arg_1 = pow(np.cos(lon) * np.cos(lat), 2)
        arg_2 = pow(np.sin(lon) * np.cos(lat), 2)
        arg_3 = pow(np.sin(lat), 2)
        vertical_norm = np.sqrt(arg_1 + arg_2 + arg_3)
        r = np.sqrt(pow(self.dx, 2) + pow(self.dy, 2) + pow(self.dz, 2))

        return np.arccos(vertical / (r * vertical_norm))

    def elevation(self, lat, lon):

        """Elevation angle fro zenital angle"""

        zangle = self.zenital_angle(lat, lon)
        return ((np.pi / 2.0) - zangle) / 0.0174533 #const.dtor

    def azimuth(self, lat, lon):

        """Azimuth angle from relative distances"""

        meridional, zonal, vertical = self.relative_directions(lat, lon)
        azimuth_angle = np.arctan2(zonal, meridional)
        if azimuth_angle < 0:
            azimuth_angle += 2.0 * np.pi
        return azimuth_angle #/ const.dtor

    def zenital_iono_angle(self, lat, lon):

        """Zenital angle projection in the ionosphere"""

        el = self.elevation(lat, lon) * dtor
       # Re = const.radius_earth
        #hm = const.avg_heigth
        return ((np.pi / 2.0) - el -
                np.arcsin((Re / (Re + hm)) * np.cos(el)))

    def coordinates(self, lat, lon):

        """ Coordinates of ionospheric piercing point (in degrees)"""

        azimuth = self.azimuth(lat, lon)
        zangle_ion = self.zenital_iono_angle(lat, lon)

        lat_ip = np.arcsin(
            np.sin(lat) * np.cos(zangle_ion) + np.cos(lat) *
                           np.sin(zangle_ion) * np.cos(azimuth)
                           )

        lon_ip = lon + np.arcsin(
            (np.sin(zangle_ion) * np.sin(azimuth)) /
                                 (np.cos(lat_ip))
                                 )

        return np.degrees(lat_ip), np.degrees(lon_ip)




import sys
import re
from datetime import datetime
import pandas as pd
import os

def SP3intp(year,doy,input_folder,output_folder):

    # Regular expressions for data analysis
    pattern_datetime = r"\*\s+(\d{4})\s+(\d{1,2})\s+(\d{1,2})\s+(\d{1,2})\s+(\d{1,2})\s+(\d{1,2}\.\d{8})"
    pattern_data = r"([A-Z]{2}\d{2})\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)"
    
    # Constructing the full path to the data folder
    path_ = input_folder
    
    # Final list for all the data (without interpolation)
    satellite_data = []
    
    # Defining the key date in dd-mm-yyyy format
    key_date = datetime.strptime(f"{doy}/{year}", "%j/%Y").strftime("%d-%m-%Y")
    
    # Check directory existence and process files if it exists
    if os.path.exists(path_):
        for file in os.listdir(path_):
            if file.endswith(".SP3"):
                file_path = os.path.join(path_, file)
                print()
                print(f"Processing file: {file_path}")
    
                # Opens the file and reads its lines
                with open(file_path, "r") as file:
                    for line in file:
                        # Extracts date and time if the line matches the pattern
                        if re.match(pattern_datetime, line):
                            match = re.match(pattern_datetime, line)
                            file_year, month, day, hour, minute = (int(match.group(i)) for i in range(1, 6))
                            second = float(match.group(6))
                            last_date_time = datetime(file_year, month, day, hour, minute, int(second))
    
                            # Skips the line if the date is greater than the key date
                            if last_date_time.date() > datetime.strptime(key_date, "%d-%m-%Y").date():
                                continue
    
                        # Checks if the line has satellite data
                        elif re.match(pattern_data, line):
                            match = re.match(pattern_data, line)
                            satellite_name, x, y, z = match.groups()
    
                            # Adds the current record to the satellite data (without interpolation)
                            satellite_data.append({
                                'Date': last_date_time.strftime("%d-%m-%Y"),
                                'Time': last_date_time.strftime("%H:%M:%S"),
                                'Satellite': satellite_name,
                                'X': x,
                                'Y': y,
                                'Z': z
                            })
    
        # Creating a DataFrame with the original data (without interpolation)
        df_satellite = pd.DataFrame(satellite_data)
        df_satellite = df_satellite.sort_values(by=['Satellite', 'Date', 'Time'])
    
        # Removes rows with dates different from the key date
        df_satellite = df_satellite[df_satellite['Date'] == key_date]
    
        # Saves the data to the file
        output_file_name = os.path.join(output_folder, f'ORBITS_{year}_{doy}.SP3')
        df_satellite.to_csv(output_file_name, sep='\t', index=False)
        print()
        print(f"Processed tabular data saved to: {output_file_name}")



from .gnss_freqs import *
from .settings import *
from .screening_settings import *
from .linear_combinations import *

from .ROTI_CALC import ROTIcalc
from .DTEC_CALC import DTECcalc
from .SIDX_CALC import SIDXcalc
from .RNX_CLEAN import RNXclean
from .SP3_INTERPOLATE import SP3intp
from .RNX_LEVELLING import RNXlevelling
from .RNX_SCREENING import RNXScreening



# linear_combinations.py

# This function calculates the Melbourne-Wubbena combination
def melbourne_wubbena(f1, f2, L1, L2, P1, P2):
    phi_lw = (f1 * L1 - f2 * L2) / (f1 - f2)
    r_pn = (f1 * P1 - f2 * P2) / (f1 - f2)
    b_w = phi_lw - r_pn
    b_w = b_w / 10**7
    return b_w




import os
import sys
import numpy as np
import pandas as pd
import numpy.ma as ma
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from datetime import datetime, timedelta
from pyOASIS import gnss_freqs

def DTECcalc(station, doy, year, input_folder, destination_directory, show_plot=True):
    """
    Compute DTEC index from GNSS data for a given station and day-of-year.

    Parameters:
        station (str): GNSS station code
        doy (int or str): Day of year
        year (int): Year of observation
        input_folder (str): Directory containing input data files
        destination_directory (str): Directory to store outputs
    """

    # Initial hour and total number of hours to process (full day)
    h1 = 0
    n_hours = 24  # hours

    # Time interval (in minutes) used in processing or averaging
    int1 = 320  # minutes

    # Get GPS frequencies from pyOASIS configuration
    gps_freqs = gnss_freqs.FREQUENCY[gnss_freqs.GPS]
    f1 = gps_freqs[1]  # L1 frequency
    f2 = gps_freqs[2]  # L2 frequency
    f5 = gps_freqs[5]  # L5 frequency

    # TEC conversion constants for L1-L2 and L1-L5 combinations (in TECU)
    akl = 40.3e16 * ((1 / f2**2) - (1 / f1**2))     # For L1-L2
    akl15 = 40.3e16 * ((1 / f5**2) - (1 / f1**2))   # For L1-L5

    # Configuration parameters for DTEC calculation

    WINDOW = 60.0 * 2.5  # Window duration in seconds (2.5 minutes)
    TDTEC = 60.0         # Time unit for expressing DTEC in TECU/min
    GAP = 1.5 * WINDOW   # Time threshold for gap detection (seconds)
    H = 450000           # Ionospheric shell height in meters
    SIGMA = 5            # Threshold multiplier for outlier removal
    DE = 3               # Degree of polynomial fit (e.g., cubic)
    GAP2 = 3600          # Minimum gap between observation arcs (seconds)
    elev_angle = 30      # Elevation cutoff angle in degrees

    # Build the full path to the station-specific subdirectory
    path_ = os.path.join(input_folder)

    # Check if the directory for the given station exists
    if os.path.exists(path_):
        # List all files in the station directory
        content_ = os.listdir(path_)

        # Filter only the relevant .RNX3 files that match the station name
        files = [file for file in content_ if file.startswith(station) and file.endswith(".RNX3")]

        # Sort the files by satellite number (assuming file name format: STATION_SAT.RNX3)
        ord_files = sorted(files, key=lambda x: int(x.split("_")[1][1:]))

        # Print the total number of valid files found
        print()
        number_of_files = len(ord_files)
        print("Number of RINEX_LEVELLING (.RNX3) files in the directory:", number_of_files)

    else:
        # If the station directory does not exist, notify the user
        print("The specified directory does not exist.")

    print()

    # Initialize empty lists to store data parsed from all .RNX3 files

    # Temporal information
    date = []                # Date string
    time2 = []               # Time string
    mjd = []                 # Modified Julian Date

    # Receiver coordinates (in ECEF, meters)
    pos_x = []               # X-coordinate
    pos_y = []               # Y-coordinate
    pos_z = []               # Z-coordinate

    # Geometry-Free combinations (used for ionospheric analysis)
    LGF_combination = []     # Geometry-Free (L1-L2) combination
    LGF_combination15 = []   # Geometry-Free (L1-L5) combination

    # Metadata
    satellites = []          # Satellite identifier (e.g., G01, R22)
    sta = []                 # GNSS station code (e.g., BOAV, LPGS)
    hght = []                # Ionospheric pierce point height (in meters)
    el = []                  # Satellite elevation angle (degrees)

    # Ionospheric pierce point (IPP) geolocation
    lonn = []                # IPP longitude (degrees)
    latt = []                # IPP latitude (degrees)

    # Raw GNSS observations
    obs_La = []              # L1 phase observation (carrier)
    obs_Lb = []              # L2 phase observation (carrier)
    obs_Lc = []              # L5 phase observation (carrier)
    obs_Ca = []              # C1 code observation (pseudorange)
    obs_Cb = []              # C2 code observation (pseudorange)
    obs_Cc = []              # C5 code observation (pseudorange)

    # Loop through each RNX3 file from the sorted list
    for file_ in ord_files:
        path_file_ = os.path.join(path_, file_)

        # Open and read the file
        with open(path_file_, 'r') as f:
            # Read the header line (tab-separated column names)
            header = f.readline().strip().split('\t')

            # Read and process each data line
            for line in f:
                # Split the line into values based on tab separator
                columns = line.strip().split('\t')

                # Build a dictionary associating column names with values
                record = {
                    'date': columns[0],
                    'time2': columns[1],
                    'mjd': columns[2],
                    'pos_x': columns[3],
                    'pos_y': columns[4],
                    'pos_z': columns[5],
                    'LGF_combination': columns[6],
                    'LGF_combination15': columns[7],
                    'satellite': columns[8],
                    'sta': columns[9],
                    'hght': columns[10],
                    'el': columns[11],
                    'lonn': columns[12],
                    'latt': columns[13],
                    'obs_La': columns[14],
                    'obs_Lb': columns[15],
                    'obs_Lc': columns[16],
                    'obs_Ca': columns[17],
                    'obs_Cb': columns[18],
                    'obs_Cc': columns[19]
                }

                # Append each field to its corresponding list
                date.append(record['date'])
                time2.append(record['time2'])
                mjd.append(record['mjd'])
                pos_x.append(record['pos_x'])
                pos_y.append(record['pos_y'])
                pos_z.append(record['pos_z'])
                LGF_combination.append(record['LGF_combination'])
                LGF_combination15.append(record['LGF_combination15'])
                satellites.append(record['satellite'])
                sta.append(record['sta'])
                hght.append(record['hght'])
                el.append(record['el'])
                lonn.append(record['lonn'])
                latt.append(record['latt'])
                obs_La.append(record['obs_La'])
                obs_Lb.append(record['obs_Lb'])
                obs_Lc.append(record['obs_Lc'])
                obs_Ca.append(record['obs_Ca'])
                obs_Cb.append(record['obs_Cb'])
                obs_Cc.append(record['obs_Cc'])


    # Create a single figure for plotting results
    plt.figure(figsize=(12, 6))  # Set the figure size (width, height in inches)

    # Define the color palette to use for different satellites
    palette = plt.get_cmap('tab10')

    # GNSS satellite system classes to analyze: 'G' for GPS, 'R' for GLONASS
    sat_classes = ['G', 'R']

    # Loop through each GNSS class (e.g., GPS, GLONASS)
    for sat in sat_classes:
        satx = sat
        print()
        print(f"Calculating DTEC for {station.upper()}  |  Year: {year}  |  DOY: {doy}")
        print()
        # Filter list of satellites that belong to the current GNSS system
        if satx:
            satellites_to_plot = [sv for sv in np.unique(satellites) if sv.startswith(satx)]
        else:
            satellites_to_plot = np.unique(satellites)

        # Initialize list to store per-satellite DTEC outputs or analysis results
        satellites_data = []

        # Process each satellite individually
        #for sat1 in satellites_to_plot:
        for idx, sat1 in enumerate(satellites_to_plot, start=1):
            print(f"Processing satellite {sat1} ({idx} of {len(satellites_to_plot)} in {satx} system)...")
            print()
            # Find all indices where the satellite ID matches sat1
            indices = np.where(np.array(satellites) == sat1)[0]

            # Initialize filtered lists to collect only the data related to this satellite
            date_filtered = []
            time2_filtered = []
            mjd_filtered = []
            pos_x_filtered = []
            pos_y_filtered = []
            pos_z_filtered = []
            LGF_combination_filtered = []
            LGF_combination15_filtered = []
            satellites_list_filtered = []
            sta_filtered = []
            hght_filtered = []
            el_filtered = []
            lonn_filtered = []
            latt_filtered = []
            obs_La_filtered = []
            obs_Lb_filtered = []
            obs_Lc_filtered = []
            obs_Ca_filtered = []
            obs_Cb_filtered = []
            obs_Cc_filtered = []

            # Loop through each selected index and extract the corresponding values
            for idx in indices:
                date_filtered.append(date[idx])
                time2_filtered.append(time2[idx])
                mjd_filtered.append(mjd[idx])
                pos_x_filtered.append(pos_x[idx])
                pos_y_filtered.append(pos_y[idx])
                pos_z_filtered.append(pos_z[idx])
                LGF_combination_filtered.append(LGF_combination[idx])
                LGF_combination15_filtered.append(LGF_combination15[idx])
                satellites_list_filtered.append(satellites[idx])
                sta_filtered.append(sta[idx])
                hght_filtered.append(hght[idx])
                el_filtered.append(el[idx])
                lonn_filtered.append(lonn[idx])
                latt_filtered.append(latt[idx])
                obs_La_filtered.append(obs_La[idx])
                obs_Lb_filtered.append(obs_Lb[idx])
                obs_Lc_filtered.append(obs_Lc[idx])
                obs_Ca_filtered.append(obs_Ca[idx])
                obs_Cb_filtered.append(obs_Cb[idx])
                obs_Cc_filtered.append(obs_Cc[idx])

            # Create a dictionary from the filtered data lists
            data = {
                'date': date_filtered,
                'time': time2_filtered,
                'mjd': mjd_filtered,
                'pos_x': pos_x_filtered,
                'pos_y': pos_y_filtered,
                'pos_z': pos_z_filtered,
                'LGF': LGF_combination_filtered,
                'LGF15': LGF_combination15_filtered,
                'satellites': satellites_list_filtered,
                'sta': sta_filtered,
                'hh': hght_filtered,
                'elev': el_filtered,
                'lonn': lonn_filtered,
                'latt': latt_filtered,
                'obs_La': obs_La_filtered,
                'obs_Lb': obs_Lb_filtered,
                'obs_Lc': obs_Lc_filtered,
                'obs_Ca': obs_Ca_filtered,
                'obs_Cb': obs_Cb_filtered,
                'obs_Cc': obs_Cc_filtered
            }

            # Convert the dictionary into a pandas DataFrame for further processing
            df = pd.DataFrame(data)

            # Create a new column combining date and time for timestamp
            df['timestamp'] = df['date'] + ' ' + df['time']

            # Convert the 'timestamp' column to pandas datetime
            df['timestamp'] = pd.to_datetime(df['timestamp'])

            # Convert key columns to float type for numerical processing
            columns_to_convert = ['LGF', 'LGF15', 'mjd', 'lonn', 'latt', 'hh', 'elev']
            df[columns_to_convert] = df[columns_to_convert].astype(float)

            # Replace invalid values (-999999.999) with NaN
            df.replace(-999999.999, np.nan, inplace=True)

            # Assign variables from DataFrame for ease of manipulation
            t = df['mjd']                          # Time in Modified Julian Date
            stec = df['LGF'] / akl                # Convert L1-L2 geometry-free to STEC (TECU)
            stec15 = df['LGF15'] / akl15          # Convert L1-L5 geometry-free to STEC
            lat = df['latt']
            lon = df['lonn']
            elev = df['elev']
            hh = df['hh']

            # Compute time differences in seconds between consecutive epochs
            d = 86400.0 * np.diff(t)

            # Determine sampling interval (e.g., 15 or 30 seconds)
            freq = int(np.ceil(np.min(d)))

            # Define time windows (T15 and T60) depending on data frequency
            if freq == 30:
                T15 = 5   # 5 samples = 2.5 minutes at 30s rate
                T60 = 20  # 20 samples = 10 minutes
            elif freq == 15:
                T15 = 15  # 15 samples = 3.75 minutes at 15s rate
                T60 = 60
            else:
                # Unexpected frequency value – set to None and print warning
                T15 = None
                T60 = None
                print(f"Unexpected frequency: {freq}")

            #print(f"T15 = {T15}, T60 = {T60}")

            # Mask zero time intervals (repeated epochs)
            d = ma.masked_values(d, 0.0)

            # Correct small STEC discontinuities (inter-frequency biases or IFB jumps)
            i = np.where(np.append(d.mask, False) == True)[0]
            for j in range(i.size):
                dIFB = stec[i[j] + 1] - stec[i[j]]
                stec[i[j] + 1:] -= dIFB

                dIFB15 = stec15[i[j] + 1] - stec15[i[j]]
                stec15[i[j] + 1:] -= dIFB15

            # Remove corresponding data points from all arrays if gaps exist
            if d.mask.any():
                mask = np.append(~d.mask, True)  # Restore size consistency
                lat = lat[mask]
                lon = lon[mask]
                hh = hh[mask]
                elev = elev[mask]
                t = t[mask]
                stec = stec[mask]
                stec15 = stec15[mask]

            # Define the starting time
            t0 = t[0]

            # Assign each epoch to a window index
            i = np.floor(np.round(86400 * (t - t0)) / WINDOW)

            # Get unique window identifiers
            j = np.unique(i)

            # Initialize accumulators for each time window
            alon = []       # Longitude
            alat = []       # Latitude
            at = []         # Window center time
            ahh = []        # Average IPP height
            DTECf = []      # DTEC L1-L2 result
            DTEC15f = []    # DTEC L1-L5 result
            elev1 = []      # Average elevation
            ROTf = []       # ROT-like parameter (optional or unused)

            # Loop over each unique window index
            for k in range(j.size):
                # Mask to select only data belonging to the current window
                l = ma.masked_values(i, j[k])

                # Only proceed if the window contains more than one observation
                if lon[l.mask].size > 1:
                    # Store averaged values for location, height, time, and elevation
                    alon.append(np.mean(lon[l.mask]))       # Average longitude
                    alat.append(np.mean(lat[l.mask]))       # Average latitude
                    ahh.append(np.mean(hh[l.mask]))         # Average IPP height
                    at.append(np.mean(t[l.mask]))           # Average time (MJD)
                    elev1.append(np.mean(elev[l.mask]))     # Average elevation angle

                    # Convert time from MJD to seconds for rolling operations
                    t_seconds = t[l.mask] * 86400.0

                    # Create a DataFrame with STEC and STEC15 time series
                    df_stec = pd.DataFrame({
                        'timestamp': t_seconds,
                        'stec': stec[l.mask],
                        'stec15': stec15[l.mask]
                    })

                    # Set time (in seconds) as the index for rolling calculations
                    df_stec.set_index('timestamp', inplace=True)

                    # Apply centered rolling average with short (T15) and long (T60) windows for L1-L2
                    stec_15min = df_stec['stec'].rolling(window=T15, min_periods=1, center=True).mean()
                    stec_60min = df_stec['stec'].rolling(window=T60, min_periods=1, center=True).mean()

                    # Apply centered rolling average for L1-L5
                    stec15_15min = df_stec['stec15'].rolling(window=T15, min_periods=1, center=True).mean()
                    stec15_60min = df_stec['stec15'].rolling(window=T60, min_periods=1, center=True).mean()

                    # Compute ΔTEC as the difference between short-term and long-term averages
                    df_stec['dtec'] = stec_15min - stec_60min
                    df_stec['dtec15'] = stec15_15min - stec15_60min

                    # Extract the most recent value from the ΔTEC series as representative
                    DTEC = df_stec['dtec'].values[-1]        # Last ΔTEC value for L1-L2
                    DTEC15 = df_stec['dtec15'].values[-1]    # Last ΔTEC value for L1-L5

                    # Append the results to the respective lists
                    DTECf.append(DTEC)
                    DTEC15f.append(DTEC15)

            # Final assignment of DTEC vectors after processing all windows
            DTEC = DTECf
            DTEC15 = DTEC15f

            ## Convert all accumulators to NumPy arrays
            alon = np.array(alon)
            alat = np.array(alat)
            ahh = np.array(ahh)
            at = np.array(at)           # Timestamps (MJD)
            DTEC = np.array(DTEC)       # ΔTEC L1-L2
            DTEC15 = np.array(DTEC15)   # ΔTEC L1-L5
            elev = np.array(elev1)      # Elevation angles

            # Compute time differences between consecutive windows (in seconds)
            d = 86400.0 * np.diff(at)

            # Mask time gaps greater than GAP2 (e.g., 3600 seconds) to identify arcs
            d = ma.masked_greater_equal(d, GAP2)

            # Indices where time gaps occur
            i = np.where(np.append(d.mask, False) == True)[0]

            # Determine start and end indices of arcs
            i1 = np.append(0, i)                         # Arc start indices
            i2 = np.append(i - 1, alon.size - 1)         # Arc end indices

            # Prepare matrices for fitted values and bounds
            y = np.empty(DTEC.size)      # Fitted ΔTEC L1-L2
            yup = np.empty(DTEC.size)    # Upper bound
            ydown = np.empty(DTEC.size)  # Lower bound

            y15 = np.empty(DTEC15.size)
            yup15 = np.empty(DTEC15.size)
            ydown15 = np.empty(DTEC15.size)

            # Polynomial fitting for each arc to smooth and detect outliers
            for j in range(i1.size):
                # Compute center time for normalization
                tm = np.mean(at[i1[j]:i2[j]+1])

                if (at[i2[j]] - at[i1[j]]) != 0.0:
                    # Normalize time over current arc
                    x = (at[i1[j]:i2[j]+1] - tm) / (at[i2[j]] - at[i1[j]])

                    # Fit polynomial of degree DE (e.g., 3)
                    c = np.polyfit(x, DTEC[i1[j]:i2[j]+1], DE)
                    c15 = np.polyfit(x, DTEC15[i1[j]:i2[j]+1], DE)

                    # Evaluate polynomial fit
                    y[i1[j]:i2[j]+1] = np.polyval(c, x)
                    y15[i1[j]:i2[j]+1] = np.polyval(c15, x)

                    # Compute RMS of residuals
                    rms = np.std(DTEC[i1[j]:i2[j]+1] - y[i1[j]:i2[j]+1])
                    rms15 = np.std(DTEC15[i1[j]:i2[j]+1] - y15[i1[j]:i2[j]+1])
                else:
                    # If arc contains only one epoch, fit is just the value itself
                    y[i1[j]:i2[j]+1] = DTEC[i1[j]:i2[j]+1]
                    y15[i1[j]:i2[j]+1] = DTEC15[i1[j]:i2[j]+1]
                    rms = 0.0
                    rms15 = 0.0

                # Define outlier detection bounds
                yup[i1[j]:i2[j]+1] = y[i1[j]:i2[j]+1] + SIGMA * rms
                ydown[i1[j]:i2[j]+1] = y[i1[j]:i2[j]+1] - SIGMA * rms

                yup15[i1[j]:i2[j]+1] = y15[i1[j]:i2[j]+1] + SIGMA * rms15
                ydown15[i1[j]:i2[j]+1] = y15[i1[j]:i2[j]+1] - SIGMA * rms15

            # Create mask for values exceeding the fit bounds (outliers)
            mask = np.abs(DTEC - y) > (yup - ydown) / 2.0

            # Remove outliers from all arrays
            alatm = alat[~mask]
            alonm = alon[~mask]
            ahhm = ahh[~mask]
            atm = at[~mask]
            DTECm = DTEC[~mask]
            DTECm15 = DTEC15[~mask]
            elevm = elev[~mask]

            # Filter out data with low elevation (below threshold)
            cutoff = np.where(elevm >= elev_angle)

            # Final filtered values (only good elevation and no outliers)
            alat = alatm[cutoff]
            alon = alonm[cutoff]
            ahh = ahhm[cutoff]
            at = atm[cutoff]
            DTEC = DTECm[cutoff]
            DTEC15 = DTECm15[cutoff]
            elev = elevm[cutoff]

            # Prepare dictionary for the current satellite's DTEC results
            satellite_data = {
                'MJD': at,
                'Longitude': alon,
                'Latitude': alat,
                'Height': ahh,
                'Elevation': elev,
                'DTEC': 10 * DTEC,       # Convert to TECU/hour
                'DTEC15': 10 * DTEC15,   # Convert to TECU/hour
                'STA': station,
                'SAT': sat1
            }

            # Store in the list of all satellites processed
            satellites_data.append(satellite_data)

        # Skip system if no data found
        if not satellites_data:
            print(f"No data found for {satx} system. Skipping...")
            continue

        # Concatenate all satellite data dictionaries into a single DataFrame
        concatenated_df = pd.concat([pd.DataFrame(data) for data in satellites_data], ignore_index=True)


        # Compute the mean values of each parameter grouped by MJD (i.e., per epoch)
        df_mean = concatenated_df.groupby('MJD').agg({
            'Longitude': 'mean',
            'Latitude': 'mean',
            'Height': 'mean',
            'Elevation': 'mean',
            'DTEC': 'mean',
            'DTEC15': 'mean',
            'STA': 'first',  # Use first occurrence for station name (constant)
            'SAT': 'first'   # Use first occurrence for satellite (may vary)
        }).reset_index()

        # Build the output file path for saving data
        output_directory = os.path.join(destination_directory)
        file_name = f"{station}_{doy}_{year}_{satx}_DTEC.txt"
        output_file_path = os.path.join(output_directory, file_name)

        # Ensure the output directory exists
        os.makedirs(output_directory, exist_ok=True)

        # Save the full dataset (tab-separated, using NaN placeholder)
        concatenated_df.to_csv(output_file_path, sep='\t', index=False, na_rep='-999999.999')

        # Convert MJD to datetime for plotting
        base_date = datetime(1858, 11, 17)
        concatenated_df['datetime'] = concatenated_df['MJD'].astype(float).apply(lambda x: base_date + timedelta(days=x))

        # Choose color and label depending on satellite system and frequency combination
        if satx == 'G':
            color_dtec = 'navy'
            label_dtec = 'DTEC: L1-L2 (GPS)'
            color_dtec15 = 'blue'
            label_dtec15 = 'DTEC: L1-L5 (GPS)'
        elif satx == 'R':
            color_dtec = 'red'
            label_dtec = 'DTEC: L1-L2 (GLONASS)'
            color_dtec15 = 'orange'
            label_dtec15 = 'DTEC: L2-L3 (GLONASS)'
        else:
            color_dtec = 'gray'
            label_dtec = f'DTEC - {satx}'
            color_dtec15 = 'darkgray'
            label_dtec15 = f'DTEC15 - {satx}'

        # Plot L1-L2 (DTEC)
        plt.scatter(concatenated_df['datetime'], concatenated_df['DTEC'], marker='o', s=30, color=color_dtec, label=label_dtec)

        # Plot L1-L5 or L2-L3 (DTEC15)
        plt.scatter(concatenated_df['datetime'], concatenated_df['DTEC15'], marker='o', s=30, color=color_dtec15, label=label_dtec15)

        # Smoothed line with gap detection (optional)
        xx = df_mean['MJD'].values
        yy = df_mean['DTEC'].values
        threshold = 0.01
        mask = np.diff(xx) > threshold
        xx_gap = np.insert(xx, np.where(mask)[0] + 1, np.nan)
        yy_gap = np.insert(yy, np.where(mask)[0] + 1, np.nan)
        datetime_gap = [base_date + timedelta(days=val) if not np.isnan(val) else np.nan for val in xx_gap]
        #plt.plot(datetime_gap, yy_gap, color='red', linewidth=2, label='Smoothed DTEC')
        # Plot smoothed DTEC line separately for GPS and GLONASS

        # Base date for datetime conversion
        base_date = datetime(1858, 11, 17)

        # Smoothed DTEC (GPS)
        df_gps = df_mean[df_mean['SAT'].str.startswith('G')]
        xx_gps = df_gps['MJD'].values
        yy_gps = df_gps['DTEC'].values
        mask_gps = np.diff(xx_gps) > 0.01
        xx_gap_gps = np.insert(xx_gps, np.where(mask_gps)[0] + 1, np.nan)
        yy_gap_gps = np.insert(yy_gps, np.where(mask_gps)[0] + 1, np.nan)
        datetime_gap_gps = [base_date + timedelta(days=val) if not np.isnan(val) else np.nan for val in xx_gap_gps]
        plt.plot(datetime_gap_gps, yy_gap_gps, color='magenta', linewidth=2, label='Smoothed DTEC (GPS)', zorder=11)

        # Smoothed DTEC (GLONASS)
        df_glonass = df_mean[df_mean['SAT'].str.startswith('R')]
        xx_glonass = df_glonass['MJD'].values
        yy_glonass = df_glonass['DTEC'].values
        mask_glonass = np.diff(xx_glonass) > 0.01
        xx_gap_glonass = np.insert(xx_glonass, np.where(mask_glonass)[0] + 1, np.nan)
        yy_gap_glonass = np.insert(yy_glonass, np.where(mask_glonass)[0] + 1, np.nan)
        datetime_gap_glonass = [base_date + timedelta(days=val) if not np.isnan(val) else np.nan for val in xx_gap_glonass]
        plt.plot(datetime_gap_glonass, yy_gap_glonass, color='gold', linewidth=2, label='Smoothed DTEC (GLONASS)', zorder=12)


        # Configure the x-axis
        hours_fmt = mdates.DateFormatter('%H')
        hour_locator = mdates.HourLocator(interval=2)
        plt.gca().xaxis.set_major_formatter(hours_fmt)
        plt.gca().xaxis.set_major_locator(hour_locator)

        # Labels and appearance
        plt.xlabel('Time (UT)', fontsize=16)
        plt.ylabel('DTEC (TECU/hour)', fontsize=16)
        plt.title(f"Station: {station.upper()}  |  Year: {year}  |  DOY: {doy}", fontsize=18)
        plt.xticks(fontsize=14)
        plt.yticks(fontsize=14)
        plt.grid(True, linestyle='--', linewidth=1, color='gray')
        # Remover duplicatas na legenda
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        plt.legend(by_label.values(), by_label.keys(), bbox_to_anchor=(1.0, 1), loc='upper left')
        plt.tight_layout()

        # Save the figure
        file_name_png = f"{station}_{doy}_{year}_DTEC.png"
        output_file_path_png = os.path.join(output_directory, file_name_png)
        plt.savefig(output_file_path_png, dpi=300)
    if show_plot:
        plt.show()
